#
# This example defines a script that solves the Traveling
# Salesperson Problem using a naive branching and bounding
# strategy. It highlights a number of advanced pybnb
# features, including:
#  (1) the use of pybnb.futures.NestedSolver
#  (2) re-continuing a solve after early termination
#
# This example can be executed in serial as
#
# $ python tsp_naive.py <data_file>
#
# or in parallel as
#
# $ mpiexec -n <n> python tsp_naive.py <data_file>
#
# The following data files are available:
# (source: https://people.sc.fsu.edu/~jburkardt/datasets/tsp/tsp.html)
#  - p01_d.txt: 15 cities, minimal tour length 291
#  - p01_d_inf.txt: same as above, but with random paths
#                   removed to make the problem infeasible
#  - fri26_d.txt: 26 cities, minimal tour length 937
#
import pybnb

def parse_dense_distance_matrix(filename):
    """Extracts a dense distance matrix from a file with
    the given name. Assumes columns are separated by
    whitespace and rows are separated by newlines. For
    consistency, entries that are zero on the off-diagonal
    will be converted to inf."""
    import math
    dist = []
    with open(filename) as f:
        line = f.readline().strip()
        while line:
            dist.append([float(x) for x in line.split()])
            line = f.readline().strip()
    N = len(dist)
    for i,row in enumerate(dist):
        assert len(row) == N
        for j,c in enumerate(row):
            assert c != -pybnb.inf
            assert not math.isnan(c)
            if i == j:
                assert c == 0
            elif c == 0:
                row[j] = pybnb.inf
    return dist

def compute_route_cost(dist, route):
    """Compute the cost of a route."""
    N = len(route)
    assert N == len(dist)+1
    assert route[0] == route[-1]
    cost = 0
    for i in range(1,len(route)):
        u = route[i-1]
        v = route[i]
        c = dist[u][v]
        assert c != 0
        cost += c
    return cost

def run_2opt(dist, cost, route):
    """Runs the 2-opt local search heuristic for TSP. Does
    not assume the distance matrix is symmetric. This
    function makes a copy of the route argument before
    running the heuristic, so the input list is not
    modified."""
    N = len(route)
    assert N == len(dist)+1
    assert route[0] == route[-1]
    route = list(route)
    while (1):
        start_over = False
        for i in range(1,N-1):
            for j in range(i+1, N):
                if j-i == 1:
                    continue
                route[i:j] = route[j-1:i-1:-1]
                new_cost = compute_route_cost(dist, route)
                if new_cost < cost:
                    cost = new_cost
                    start_over = True
                    break
                else:
                    # reset the route
                    route[i:j] = route[j-1:i-1:-1]
            if start_over:
                break
        if start_over:
            continue
        break
    return cost, route

class TSP_Naive(pybnb.Problem):

    def __init__(self, dist):
        self._dist = dist
        self._N = len(dist)
        # state that changes during the solve
        self._path = [0]

    #
    # Implement Problem abstract methods
    #

    def sense(self):
        return pybnb.minimize

    def objective(self):
        cost = self.infeasible_objective()
        if len(self._path) == self._N:
            return_cost = self._dist[self._path[-1]][self._path[0]]
            assert return_cost != 0
            if return_cost != pybnb.inf:
                cost = 0.0
                for i in range(self._N-1):
                    cost += self._dist[self._path[i]][self._path[i+1]]
                cost += return_cost
        return cost

    def bound(self):
        if len(self._path) == self._N:
            return self.objective()

        visited = set(self._path)
        remaining = []
        for u in range(self._N):
            if u not in visited:
                remaining.append(u)
        assert len(remaining) > 0

        bound = 0
        # for the edges that are certain
        for i in range(len(self._path) - 1):
            bound += self._dist[self._path[i]][self._path[i+1]]
        # for the last item
        last = self._path[-1]
        tmp = [self._dist[last][v] for v in remaining
               if ((self._dist[last][v] != pybnb.inf) and \
                   (v != last))]
        if len(tmp) == 0:
            return self.infeasible_objective()
        bound += min(tmp)
        # for the undetermined nodes
        p = [self._path[0]] + remaining
        for r in remaining:
            tmp = [self._dist[r][v] for v in p
                   if ((self._dist[r][v] != pybnb.inf) and \
                       (v != r))]
            if len(tmp) == 0:
                return self.infeasible_objective()
            bound += min(tmp)
        return bound

    def save_state(self, node):
        node.state = self._path

    def load_state(self, node):
        self._path = node.state
        assert len(self._path) <= self._N

    def branch(self):
        # note that the branch method should never be called
        # with a path of length N as the objective and bound
        # converge exactly in that case.
        assert len(self._path) < self._N
        u = self._path[-1]
        visited = set(self._path)
        for v in range(self._N):
            # dist[u][v] == inf means no edge
            if (self._dist[u][v] != pybnb.inf) and \
               (v not in visited):
                assert self._dist[u][v] != 0
                child = pybnb.Node()
                child.state = self._path + [v]
                yield child

    def notify_solve_finished(self,
                              comm,
                              worker_comm,
                              results):
        tour = None
        if (results.best_node is not None) and \
           (results.best_node.state is not None):
            path = results.best_node.state
            route = path + [path[0]]
            tour = {'cost': results.best_node.objective,
                    'route': route}
        results.tour = tour

if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description=("Run parallel branch and bound "
                     "to solve an instance of TSP."))
    parser.add_argument("data_filename", type=str,
                        help=("The name of a file that stores a "
                              "dense distance matrix."))
    parser.add_argument("--results-filename", type=str, default=None,
                        help=("When set, saves the solver results "
                              "into a YAML-formatted file with the "
                              "given name."))
    args = parser.parse_args()

    dist = parse_dense_distance_matrix(args.data_filename)
    problem = TSP_Naive(dist)
    solver = pybnb.Solver()

    # The following solve loop does the following:
    #  (1) Solve the tsp problem using a nested
    #      branch-and-bound strategy until any improvement
    #      to the previous best cost is made (objective_stop)
    #  (2) If the solution status from (1) is feasible, run
    #      the 2-opt heuristic to attempt to improve the
    #      solution. For any other solution status (e.g.,
    #      optimal, infeasible), exit the solve loop.
    #  (3) Go to step (1), initializing the solve with the
    #      remaining queue items from the previous solve
    #      (initialize_queue), a potentially new best node
    #      created with the solution returned from the 2-opt
    #      heuristic (best_node), and a new objective_stop
    #      value of one less than the current best cost (so
    #      we can go to step (2) if a new solution is
    #      found).
    objective_stop = pybnb.inf
    queue = None
    best_node = None
    while (1):
        # solve exactly (disable check for relative
        # gap and use absolute gap of zero)
        results = solver.solve(
            pybnb.futures.NestedSolver(problem,
                                       queue_strategy='depth',
                                       time_limit=1),
            absolute_gap=0,
            relative_gap=None,
            queue_strategy='depth',
            initialize_queue=queue,
            best_node=best_node,
            objective_stop=objective_stop)
        if (results.solution_status == "feasible") and \
           (results.termination_condition != "interrupted"):
           assert results.best_node is not None
           assert results.tour is not None
           cost, route = run_2opt(dist,
                                  results.tour['cost'],
                                  results.tour['route'])
           if cost < results.tour['cost']:
               if solver.is_dispatcher:
                   print("Local heuristic improved best tour:")
                   print(" -  cost: "+str(cost))
                   print(" - route: "+str(route))
           best_node = pybnb.Node()
           best_node.objective = cost
           best_node.state = route[:-1]
           objective_stop = cost - 1
           queue = solver.save_dispatcher_queue()
        else:
            if solver.is_dispatcher:
                print("Terminating the solve loop.")
                print("Final solution status: "
                      +str(results.solution_status))
            break

    stats = solver.collect_worker_statistics()
    if solver.is_dispatcher:
        pybnb.solver.summarize_worker_statistics(stats)
        # save results to a file
        # (mainly used for testing this example)
        if args.results_filename is not None:
            results.write(args.results_filename)



import pytest

from pybnb.common import (minimize,
                          maximize,
                          inf)
from pybnb.misc import get_simple_logger
from pybnb.node import Node
from pybnb.convergence_checker import ConvergenceChecker
from pybnb.dispatcher import (DispatcherLocal,
                              DispatcherQueueData)
from pybnb.priority_queue import (WorstBoundFirstPriorityQueue,
                                  CustomPriorityQueue,
                                  BestObjectiveFirstPriorityQueue,
                                  BreadthFirstPriorityQueue,
                                  DepthFirstPriorityQueue,
                                  FIFOQueue,
                                  LIFOQueue,
                                  RandomPriorityQueue,
                                  LocalGapPriorityQueue,
                                  LexicographicPriorityQueue)

class TestDispatcherQueueData(object):

    def test_bound(self):

        for sense in (minimize, maximize):
            queue = DispatcherQueueData([],
                                        None,
                                        sense)
            assert queue.nodes == []
            assert queue.worst_terminal_bound is None
            assert queue.sense == sense
            assert queue.bound() is None
            queue = DispatcherQueueData([],
                                        0,
                                        sense)
            assert queue.nodes == []
            assert queue.worst_terminal_bound == 0
            assert queue.sense == sense
            assert queue.bound() == 0
            queue = DispatcherQueueData([Node()],
                                        0,
                                        sense)
            queue.nodes[0].bound = (-1 if (sense == minimize) else 1)
            assert len(queue.nodes) == 1
            assert queue.worst_terminal_bound == 0
            assert queue.sense == sense
            assert queue.bound() == (-1 if (sense == minimize) else 1)
            queue.worst_terminal_bound = \
                (-2 if (sense == minimize) else 2)
            assert queue.bound() == (-2 if (sense == minimize) else 2)

class TestDispatcherSimple(object):

    def test_queue_strategy(self):
        node_limit = None
        time_limit = None
        log = get_simple_logger()
        log_interval_seconds = inf
        log_new_incumbent = True
        convergence_checker = ConvergenceChecker(minimize)

        root = Node()
        root.tree_depth = 0
        root.bound = convergence_checker.unbounded_objective
        root.objective = convergence_checker.infeasible_objective
        queue = DispatcherQueueData(
            [root],
            None,
            minimize)

        disp = DispatcherLocal()
        disp.initialize(
            convergence_checker.infeasible_objective,
            None,
            queue,
            'bound',
            convergence_checker,
            node_limit,
            time_limit,
            log,
            log_interval_seconds,
            log_new_incumbent)
        assert type(disp.queue) is WorstBoundFirstPriorityQueue
        disp.initialize(
            convergence_checker.infeasible_objective,
            None,
            queue,
            'custom',
            convergence_checker,
            node_limit,
            time_limit,
            log,
            log_interval_seconds,
            log_new_incumbent)
        assert type(disp.queue) is CustomPriorityQueue
        disp.initialize(
            convergence_checker.infeasible_objective,
            None,
            queue,
            'objective',
            convergence_checker,
            node_limit,
            time_limit,
            log,
            log_interval_seconds,
            log_new_incumbent)
        assert type(disp.queue) is BestObjectiveFirstPriorityQueue
        disp.initialize(
            convergence_checker.infeasible_objective,
            None,
            queue,
            'breadth',
            convergence_checker,
            node_limit,
            time_limit,
            log,
            log_interval_seconds,
            log_new_incumbent)
        assert type(disp.queue) is BreadthFirstPriorityQueue
        disp.initialize(
            convergence_checker.infeasible_objective,
            None,
            queue,
            'depth',
            convergence_checker,
            node_limit,
            time_limit,
            log,
            log_interval_seconds,
            log_new_incumbent)
        assert type(disp.queue) is DepthFirstPriorityQueue
        disp.initialize(
            convergence_checker.infeasible_objective,
            None,
            queue,
            'fifo',
            convergence_checker,
            node_limit,
            time_limit,
            log,
            log_interval_seconds,
            log_new_incumbent)
        assert type(disp.queue) is FIFOQueue
        disp.initialize(
            convergence_checker.infeasible_objective,
            None,
            queue,
            'lifo',
            convergence_checker,
            node_limit,
            time_limit,
            log,
            log_interval_seconds,
            log_new_incumbent)
        assert type(disp.queue) is LIFOQueue
        disp.initialize(
            convergence_checker.infeasible_objective,
            None,
            queue,
            'random',
            convergence_checker,
            node_limit,
            time_limit,
            log,
            log_interval_seconds,
            log_new_incumbent)
        assert type(disp.queue) is RandomPriorityQueue
        disp.initialize(
            convergence_checker.infeasible_objective,
            None,
            queue,
            'local_gap',
            convergence_checker,
            node_limit,
            time_limit,
            log,
            log_interval_seconds,
            log_new_incumbent)
        assert type(disp.queue) is LocalGapPriorityQueue
        disp.initialize(
            convergence_checker.infeasible_objective,
            None,
            queue,
            ('bound','local_gap'),
            convergence_checker,
            node_limit,
            time_limit,
            log,
            log_interval_seconds,
            log_new_incumbent)
        assert type(disp.queue) is LexicographicPriorityQueue

    def test_initialize_queue(self):
        node_limit = None
        time_limit = None
        log = get_simple_logger()
        log_interval_seconds = inf
        log_new_incumbent = True
        convergence_checker = ConvergenceChecker(minimize)

        root = Node()
        root.tree_depth = 0
        root.bound = convergence_checker.unbounded_objective
        root.objective = convergence_checker.infeasible_objective
        queue = DispatcherQueueData(
            [root],
            None,
            minimize)

        best_node_ = Node()
        best_node_.objective = 0
        best_node_._generate_uuid()
        disp = DispatcherLocal()
        disp.initialize(
            convergence_checker.infeasible_objective,
            best_node_,
            queue,
            'bound',
            convergence_checker,
            node_limit,
            time_limit,
            log,
            log_interval_seconds,
            log_new_incumbent)
        assert disp.best_objective == 0
        assert disp.best_node.objective == 0
        assert disp.best_node is best_node_
        disp.initialize(
            -1,
            best_node_,
            queue,
            'bound',
            convergence_checker,
            node_limit,
            time_limit,
            log,
            log_interval_seconds,
            log_new_incumbent)
        assert disp.best_objective == -1
        assert disp.best_node.objective == 0
        assert disp.best_node is best_node_
        best_node_.objective = 1
        disp.initialize(
            2,
            best_node_,
            queue,
            'bound',
            convergence_checker,
            node_limit,
            time_limit,
            log,
            log_interval_seconds,
            log_new_incumbent)
        assert disp.best_objective == 1
        assert disp.best_node.objective == 1
        assert disp.best_node is best_node_
        best_node_.objective = 1
        root.objective = -1
        disp.initialize(
            2,
            best_node_,
            queue,
            'bound',
            convergence_checker,
            node_limit,
            time_limit,
            log,
            log_interval_seconds,
            log_new_incumbent)
        assert disp.best_objective == 1
        assert disp.best_node.objective == 1
        assert disp.best_node is best_node_
        best_node_.objective = 1
        root.objective = -1
        disp.initialize(
            -2,
            best_node_,
            queue,
            'bound',
            convergence_checker,
            node_limit,
            time_limit,
            log,
            log_interval_seconds,
            log_new_incumbent)
        assert disp.best_objective == -2
        assert disp.best_node.objective == 1
        assert disp.best_node is best_node_
        # bad objective sense
        queue = DispatcherQueueData(
            [root],
            None,
            maximize)
        with pytest.raises(ValueError):
            disp.initialize(
                convergence_checker.infeasible_objective,
                best_node_,
                queue,
                'bound',
                convergence_checker,
                node_limit,
                time_limit,
                log,
                log_interval_seconds,
                log_new_incumbent)



#
# This example solves the binary knapsack problem.
#

import pybnb

class BinaryKnapsack(pybnb.Problem):

    def __init__(self, W, v, w):
        assert W >= 0
        assert len(v) == len(w)
        self._W = W
        self._v = list(v)
        self._w = list(w)
        self._n = len(self._v)
        self._sorted_order = sorted(
            range(self._n),
            key=lambda i: self._v[i]/float(self._w[i]),
            reverse=True)
        self._weight = 0
        self._value = 0
        self._level = 0
        self._choices = []

    #
    # Implement Problem abstract methods
    #

    def sense(self):
        return pybnb.maximize

    def objective(self):
        return self._value

    def bound(self):
        weight = self._weight
        bound = self._value
        added = False
        for level in range(self._level, self._n):
            k = self._sorted_order[level]
            next_weight = weight + self._w[k]
            if next_weight > self._W:
                break
            weight = next_weight
            bound += self._v[k]
            added = True
        else: # no break in for-loop
            # there are no items left
            # to add partially
            return bound
        if not added:
            # terminal node
            return self.objective()
        return bound + \
            (self._W - weight)*(self._v[k]/float(self._w[k]))

    def save_state(self, node):
        node.state = (self._weight,
                      self._value,
                      self._level,
                      self._choices)

    def load_state(self, node):
        (self._weight,
         self._value,
         self._level,
         self._choices) = node.state
        assert len(self._choices) <= self._n
        assert self._weight <= self._W
        assert self._level <= self._n

    def branch(self):
        assert len(self._choices) < self._n
        for level in range(self._level, self._n):
            i = self._sorted_order[level]
            child_weight = self._weight + self._w[i]
            if child_weight <= self._W:
                child_value = self._value + self._v[i]
                child = pybnb.Node()
                # we know the child objective value, so
                # assign it to the child node rather than
                # letting it inherit the parent objective
                # value (this may be useful for queue
                # prioritization)
                child.objective = child_value
                child.state = (child_weight,
                               child_value,
                               level + 1,
                               self._choices + [i])
                yield child

if __name__ == "__main__":
    import pybnb.misc

    W = 25
    w = [2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2]
    v = [2.05,2.06,2.07,2.08,2.09,2.10,2.11,
         2.12,2.13,2.14,2.15,2.16,2.17,2.18,
         2.19,2.20,2.21,2.22,2.23,2.24]

    problem = BinaryKnapsack(W,v,w)
    pybnb.misc.create_command_line_solver(problem)



import sys
from runtests.mpi import Tester
import os.path
# code is placed inside this block so that it is not
# executed when imported during doctests
if __name__ == "__main__":
    tester = Tester(os.path.join(os.path.abspath(__file__)),
                "../../src", # tests are executed from within ./build/tests
                extra_path=None)
    tester.main(sys.argv[1:])



import pybnb

class DelayedUnboundedMax(pybnb.Problem):

    #
    # Implement Problem abstract methods
    #

    def sense(self):
        return pybnb.maximize
    def objective(self):
        if self._node.tree_depth > 1:
            return self.unbounded_objective()
        else:
            return 0
    def bound(self):
        return self.unbounded_objective()
    def save_state(self, node):
        pass
    def load_state(self, node):
        self._node = node
    def branch(self):
        yield pybnb.Node()
        yield pybnb.Node()
        yield pybnb.Node()
        yield pybnb.Node()
        yield pybnb.Node()



"""
A collection of priority queue implementations that can be
used by the dispatcher.

Copyright by Gabriel A. Hackebeil (gabe.hackebeil@gmail.com).
"""
import random
import collections
import heapq
import math

from pybnb.common import (minimize,
                          maximize)

from sortedcontainers import SortedList
import six

class _NoThreadingMaxPriorityFirstQueue(object):
    """A simple priority queue implementation that is not
    thread safe. When the queue is not empty, the item with
    the highest priority is next.

    This queue implementation is not allowed to store None.
    """
    requires_priority = True

    def __init__(self):
        self._count = 0
        self._heap = []

    def _negate(self, priority):
        if hasattr(priority, '__neg__'):
            return -priority
        else:
            return tuple(-v for v in priority)

    def size(self):
        """Returns the size of the queue."""
        return len(self._heap)

    def put(self, item, priority, _push_=heapq.heappush):
        """Puts an item into the queue with the given
        priority. Items placed in the queue may not be
        None. This method returns a unique counter associated
        with each put."""
        if item is None:
            raise ValueError("queue item can not be None")
        cnt = self._count
        self._count += 1
        _push_(self._heap, (self._negate(priority), cnt, item))
        return cnt

    def get(self, _pop_=heapq.heappop):
        """Removes and returns the highest priority item in
        the queue, where ties are broken by the order items
        were placed in the queue. If the queue is empty,
        returns None."""
        if len(self._heap) > 0:
            return _pop_(self._heap)[2]
        else:
            return None

    def put_get(self, item, priority, _push_pop_=heapq.heappushpop):
        """Combines a put and get call, which can be more
        efficient than two separate put and get
        calls. Returns a tuple containing the put and get
        return values."""
        if item is None:
            raise ValueError("queue item can not be None")
        cnt = self._count
        self._count += 1
        if len(self._heap) > 0:
            item_ = _push_pop_(self._heap,
                               (self._negate(priority), cnt, item))[2]
            return cnt, item_
        else:
            return cnt, item

    def next(self):
        """Returns, without modifying the queue, a tuple of
        the form (cnt, item), where item is highest priority
        entry in the queue and cnt is the unique counter
        assigned to it when it was added to the queue.

        Raises
        ------
        IndexError
            If the queue is empty.
        """
        try:
            return self._heap[0][1:]
        except IndexError:
            raise IndexError("The queue is empty")

    def filter(self, func, include_counters=False):
        """Removes items from the queue for which
        `func(item)` returns False. The list of items
        removed is returned. If `include_counters` is set to
        True, values in the returned list will have the form
        (cnt, item), where cnt is a unique counter that was
        created for the item when it was added to the
        queue."""
        heap_new = []
        removed = []
        for priority, cnt, item in self._heap:
            if func(item):
                heap_new.append((priority, cnt, item))
            elif not include_counters:
                removed.append(item)
            else:
                removed.append((cnt,item))
        heapq.heapify(heap_new)
        self._heap = heap_new
        return removed

    def items(self):
        """Iterates over the queued items in arbitrary order
        without modifying the queue."""
        for _,_,item in self._heap:
            yield item

class _NoThreadingFIFOQueue(object):
    """A simple first-in, first-out queue implementation
    that is not thread safe.

    This queue implementation is not allowed to store None.
    """
    requires_priority = False

    def __init__(self):
        self._count = 0
        self._deque = collections.deque()

    def size(self):
        """Returns the size of the queue."""
        return len(self._deque)

    def put(self, item):
        """Puts an item into the queue. Items placed in the
        queue may not be None. This method returns a unique
        counter associated with each put."""
        if item is None:
            raise ValueError("queue item can not be None")
        cnt = self._count
        self._count += 1
        self._deque.append((cnt, item))
        return cnt

    def get(self):
        """Removes and returns the next item in the
        queue. If the queue is empty, returns None."""
        if len(self._deque) > 0:
            return self._deque.popleft()[1]
        else:
            return None

    def put_get(self, item):
        """Combines a put and get call, which can be more
        efficient than two separate put and get
        calls. Returns a tuple containing the put and get
        return values."""
        if item is None:
            raise ValueError("queue item can not be None")
        cnt = self._count
        self._count += 1
        if len(self._deque) > 0:
            self._deque.rotate(-1)
            return_item = self._deque[-1][1]
            self._deque[-1] = (cnt, item)
            return cnt, return_item
        else:
            return cnt, item

    def next(self):
        """Returns, without modifying the queue, a tuple of
        the form (cnt, item), where item is highest priority
        entry in the queue and cnt is the unique counter
        assigned to it when it was added to the queue.

        Raises
        ------
        IndexError
            If the queue is empty.
        """
        try:
            return self._deque[0]
        except IndexError:
            raise IndexError("The queue is empty")

    def filter(self, func, include_counters=False):
        """Removes items from the queue for which
        `func(item)` returns False. The list of items
        removed is returned. If `include_counters` is set to
        True, values in the returned list will have the form
        (cnt, item), where cnt is a unique counter that was
        created for the item when it was added to the
        queue."""
        deque_new = collections.deque()
        removed = []
        for cnt, item in self._deque:
            if func(item):
                deque_new.append((cnt, item))
            elif not include_counters:
                removed.append(item)
            else:
                removed.append((cnt,item))
        self._deque = deque_new
        return removed

    def items(self):
        """Iterates over the queued items in arbitrary order
        without modifying the queue."""
        for _,item in self._deque:
            yield item

class _NoThreadingLIFOQueue(object):
    """A simple last-in, first-out queue implementation
    that is not thread safe.

    This queue implementation is not allowed to store None.
    """
    requires_priority = False

    def __init__(self):
        self._count = 0
        self._items = []

    def size(self):
        """Returns the size of the queue."""
        return len(self._items)

    def put(self, item):
        """Puts an item into the queue. Items placed in the
        queue may not be None. This method returns a unique
        counter associated with each put."""
        if item is None:
            raise ValueError("queue item can not be None")
        cnt = self._count
        self._count += 1
        self._items.append((cnt, item))
        return cnt

    def get(self):
        """Removes and returns the next item in the
        queue. If the queue is empty, returns None."""
        if len(self._items) > 0:
            return self._items.pop()[1]
        else:
            return None

    def put_get(self, item):
        """Combines a put and get call, which can be more
        efficient than two separate put and get
        calls. Returns a tuple containing the put and get
        return values."""
        if item is None:
            raise ValueError("queue item can not be None")
        cnt = self._count
        self._count += 1
        return cnt, item

    def next(self):
        """Returns, without modifying the queue, a tuple of
        the form (cnt, item), where item is highest priority
        entry in the queue and cnt is the unique counter
        assigned to it when it was added to the queue.

        Raises
        ------
        IndexError
            If the queue is empty.
        """
        try:
            return self._items[-1]
        except IndexError:
            raise IndexError("The queue is empty")

    def filter(self, func, include_counters=False):
        """Removes items from the queue for which
        `func(item)` returns False. The list of items
        removed is returned. If `include_counters` is set to
        True, values in the returned list will have the form
        (cnt, item), where cnt is a unique counter that was
        created for the item when it was added to the
        queue."""
        items_new = []
        removed = []
        for cnt, item in self._items:
            if func(item):
                items_new.append((cnt, item))
            elif not include_counters:
                removed.append(item)
            else:
                removed.append((cnt,item))
        self._items = items_new
        return removed

    def items(self):
        """Iterates over the queued items in arbitrary order
        without modifying the queue."""
        for _,item in self._items:
            yield item

class IPriorityQueue(object):
    """The abstract interface for priority queues that store
    node data for the dispatcher."""

    def size(self):                               #pragma:nocover
        """Returns the size of the queue."""
        raise NotImplementedError

    def put(self, node):                          #pragma:nocover
        """Puts an node in the queue, possibly updating the
        value of :attr:`queue_priority <pybnb.node.Node.queue_priority>`,
        depending on the queue implementation. This method
        returns a unique counter associated with each
        put."""
        raise NotImplementedError()

    def get(self):                                #pragma:nocover
        """Returns the next node in the queue. If the queue
        is empty, returns None."""
        raise NotImplementedError()

    def bound(self):                              #pragma:nocover
        """Returns the weakest bound of all nodes in the
        queue. If the queue is empty, returns None."""
        raise NotImplementedError()

    def filter(self, func):                       #pragma:nocover
        """Removes nodes from the queue for which
        `func(node)` returns False. The list of nodes
        removed is returned. If the queue is empty or no
        nodes are removed, the returned list will be
        empty."""
        raise NotImplementedError()

    def items(self):                              #pragma:nocover
        """Iterates over the queued nodes in arbitrary order
        without modifying the queue."""
        raise NotImplementedError()

class WorstBoundFirstPriorityQueue(IPriorityQueue):
    """A priority queue implementation that serves nodes
    with the worst bound first.

    Parameters
    ----------
    sense : {:obj:`minimize <pybnb.common.minimize>`, :obj:`maximize <pybnb.common.maximize>`}
        The objective sense for the problem.
    """

    def __init__(self, sense):
        assert sense in (minimize, maximize)
        self._sense = sense
        self._queue = _NoThreadingMaxPriorityFirstQueue()

    @staticmethod
    def generate_priority(node, sense, queue):
        bound = node.bound
        assert not math.isnan(bound)
        if sense == minimize:
            return -bound
        else:
            assert sense == maximize
            return bound

    def size(self):
        return self._queue.size()

    def put(self, node):
        node.queue_priority = self.generate_priority(node,
                                                     self._sense,
                                                     None)
        return self._queue.put(node, node.queue_priority)

    def get(self):
        return self._queue.get()

    def bound(self):
        try:
            return self._queue.next()[1].bound
        except IndexError:
            return None

    def filter(self, func):
        return self._queue.filter(func)

    def items(self):
        return self._queue.items()

class CustomPriorityQueue(IPriorityQueue):
    """A priority queue implementation that can handle
    custom node priorities. It uses an additional data
    structure to reduce the amount of time it takes to
    compute a queue bound.

    Parameters
    ----------
    sense : {:obj:`minimize <pybnb.common.minimize>`, :obj:`maximize <pybnb.common.maximize>`}
        The objective sense for the problem.
    """

    def __init__(self,
                 sense,
                 _queue_type_=_NoThreadingMaxPriorityFirstQueue):
        assert sense in (minimize, maximize)
        self._sense = sense
        self._queue = _queue_type_()
        self._sorted_by_bound = SortedList()

    def size(self):
        return self._queue.size()

    def put(self, node):
        bound = node.bound
        assert not math.isnan(bound)
        if self._queue.requires_priority:
            priority = node.queue_priority
            if priority is None:
                raise ValueError("A node queue priority is required")
            cnt = self._queue.put(node, priority)
        else:
            cnt = self._queue.put(node)
        if self._sense == maximize:
            self._sorted_by_bound.add((-bound, cnt, node))
        else:
            self._sorted_by_bound.add((bound, cnt, node))
        return cnt

    def get(self):
        if self._queue.size() > 0:
            cnt, tmp_ = self._queue.next()
            assert type(cnt) is int
            node = self._queue.get()
            assert tmp_ is node
            bound = node.bound
            if self._sense == maximize:
                self._sorted_by_bound.remove((-bound, cnt, node))
            else:
                self._sorted_by_bound.remove((bound, cnt, node))
            return node
        else:
            return None

    def bound(self):
        try:
            return self._sorted_by_bound[0][2].bound
        except IndexError:
            return None

    def filter(self, func):
        removed = []
        for cnt, node in self._queue.filter(func,
                                            include_counters=True):
            removed.append(node)
            bound = node.bound
            if self._sense == maximize:
                self._sorted_by_bound.remove((-bound, cnt, node))
            else:
                self._sorted_by_bound.remove((bound, cnt, node))
        return removed

    def items(self):
        return self._queue.items()

class BestObjectiveFirstPriorityQueue(CustomPriorityQueue):
    """A priority queue implementation that serves nodes
    with the best objective first.

    sense : {:obj:`minimize <pybnb.common.minimize>`, :obj:`maximize <pybnb.common.maximize>`}
        The objective sense for the problem.
    """

    @staticmethod
    def generate_priority(node, sense, queue):
        objective = node.objective
        assert not math.isnan(objective)
        if sense == minimize:
            return -objective
        else:
            assert sense == maximize
            return objective

    def put(self, node):
        node.queue_priority = self.generate_priority(node,
                                                     self._sense,
                                                     None)
        return super(BestObjectiveFirstPriorityQueue, self).put(node)

class BreadthFirstPriorityQueue(CustomPriorityQueue):
    """A priority queue implementation that serves nodes in
    breadth-first order.

    sense : {:obj:`minimize <pybnb.common.minimize>`, :obj:`maximize <pybnb.common.maximize>`}
        The objective sense for the problem.
    """

    @staticmethod
    def generate_priority(node, sense, queue):
        assert node.tree_depth >= 0
        return -node.tree_depth

    def put(self, node):
        node.queue_priority = self.generate_priority(node,
                                                     None,
                                                     None)
        return super(BreadthFirstPriorityQueue, self).put(node)

class DepthFirstPriorityQueue(CustomPriorityQueue):
    """A priority queue implementation that serves nodes in
    depth-first order.

    sense : {:obj:`minimize <pybnb.common.minimize>`, :obj:`maximize <pybnb.common.maximize>`}
        The objective sense for the problem.
    """

    @staticmethod
    def generate_priority(node, sense, queue):
        assert node.tree_depth >= 0
        return node.tree_depth

    def put(self, node):
        node.queue_priority = self.generate_priority(node,
                                                     None,
                                                     None)
        return super(DepthFirstPriorityQueue, self).put(node)

class FIFOQueue(CustomPriorityQueue):
    """A priority queue implementation that serves nodes in
    first-in, first-out order.

    sense : {:obj:`minimize <pybnb.common.minimize>`, :obj:`maximize <pybnb.common.maximize>`}
        The objective sense for the problem.
    """

    def __init__(self, sense):
        super(FIFOQueue, self).__init__(
            sense,
            _queue_type_=_NoThreadingFIFOQueue)

    @staticmethod
    def generate_priority(node, sense, queue):
        return -queue._count

    def put(self, node):
        node.queue_priority = self.generate_priority(None,
                                                     None,
                                                     self._queue)
        cnt = super(FIFOQueue, self).put(node)
        assert node.queue_priority == -cnt
        return cnt

class LIFOQueue(CustomPriorityQueue):
    """A priority queue implementation that serves nodes in
    last-in, first-out order.

    sense : {:obj:`minimize <pybnb.common.minimize>`, :obj:`maximize <pybnb.common.maximize>`}
        The objective sense for the problem.
    """

    def __init__(self, sense):
        super(LIFOQueue, self).__init__(
            sense,
            _queue_type_=_NoThreadingLIFOQueue)

    @staticmethod
    def generate_priority(node, sense, queue):
        return queue._count

    def put(self, node):
        node.queue_priority = self.generate_priority(None,
                                                     None,
                                                     self._queue)
        cnt = super(LIFOQueue, self).put(node)
        assert node.queue_priority == cnt
        return cnt

class RandomPriorityQueue(CustomPriorityQueue):
    """A priority queue implementation that assigns
    a random priority to each incoming node.

    sense : {:obj:`minimize <pybnb.common.minimize>`, :obj:`maximize <pybnb.common.maximize>`}
        The objective sense for the problem.
    """

    @staticmethod
    def generate_priority(node, sense, queue):
        return random.random()

    def put(self, node):
        node.queue_priority = self.generate_priority(None,None,None)
        return super(RandomPriorityQueue, self).put(node)

class LocalGapPriorityQueue(CustomPriorityQueue):
    """A priority queue implementation that serves nodes
    with the largest gap between the local objective and
    bound first.

    sense : {:obj:`minimize <pybnb.common.minimize>`, :obj:`maximize <pybnb.common.maximize>`}
        The objective sense for the problem.
    """

    @staticmethod
    def generate_priority(node, sense, queue):
        objective = node.objective
        bound = node.bound
        if sense == minimize:
            gap = objective - bound
        else:
            assert sense == maximize
            gap = bound - objective
        assert not math.isnan(gap)
        return gap

    def put(self, node):
        node.queue_priority = self.generate_priority(node,
                                                     self._sense,
                                                     None)
        return super(LocalGapPriorityQueue, self).put(node)

class LexicographicPriorityQueue(CustomPriorityQueue):
    """A priority queue implementation that serves nodes
    with the largest gap between the local objective and
    bound first.

    sense : {:obj:`minimize <pybnb.common.minimize>`, :obj:`maximize <pybnb.common.maximize>`}
        The objective sense for the problem.
    """

    def __init__(self, queue_types, sense):
        self._queue_types = tuple(queue_types)
        assert len(self._queue_types)
        super(LexicographicPriorityQueue, self).__init__(sense)

    def _generate_priority(self, node):
        return tuple(qt.generate_priority(node,
                                          self._sense,
                                          self._queue)
                     for qt in self._queue_types)

    def put(self, node):
        node.queue_priority = self._generate_priority(node)
        return super(LexicographicPriorityQueue, self).put(node)

def PriorityQueueFactory(name, *args, **kwds):
    """Returns a new instance of the priority queue type
    registered under the given name."""
    if isinstance(name, six.string_types):
        if name not in PriorityQueueFactory._types:
            raise ValueError("invalid queue type: %s"
                             % (name))
        return PriorityQueueFactory._types[name](*args,
                                                 **kwds)
    else:
        names = []
        for n_ in name:
            print(n_)
            if n_ not in PriorityQueueFactory._types:
                raise ValueError("invalid queue type: %s"
                                 % (n_))
            if n_ == 'custom':
                raise ValueError("'custom' queue type not "
                                 "allowed when defining a "
                                 "lexicographic queue strategy")
            names.append(PriorityQueueFactory._types[n_])
        if len(names) == 0:
            raise ValueError("Can not define lexicographic queue "
                             "strategy with empty list")
        return LexicographicPriorityQueue(names, *args, **kwds)
PriorityQueueFactory._types = {}

def register_queue_type(name, cls):
    """Registers a new priority queue class with the
    PriorityQueueFactory."""
    if (name in PriorityQueueFactory._types) and \
       (PriorityQueueFactory._types[name] is not cls):
        raise ValueError(
            "The name '%s' has already been registered"
            "for priority queue type '%s'"
            % (name, cls))
    PriorityQueueFactory._types[name] = cls

register_queue_type('bound',
                    WorstBoundFirstPriorityQueue)
register_queue_type('custom',
                    CustomPriorityQueue)
register_queue_type('objective',
                    BestObjectiveFirstPriorityQueue)
register_queue_type('breadth',
                    BreadthFirstPriorityQueue)
register_queue_type('depth',
                    DepthFirstPriorityQueue)
register_queue_type('fifo',
                    FIFOQueue)
register_queue_type('lifo',
                    LIFOQueue)
register_queue_type('random',
                    RandomPriorityQueue)
register_queue_type('local_gap',
                    LocalGapPriorityQueue)



import pytest

import pybnb

pyomo_available = False
try:
    import pyomo.kernel as pmo
    from pybnb.pyomo.misc import (add_tmp_component,
                                  create_optimality_bound,
                                  correct_integer_lb,
                                  correct_integer_ub)
    from pybnb.pyomo.problem import PyomoProblem
    from pybnb.pyomo.range_reduction import RangeReductionProblem
    pyomo_available = True
except:
    pass

class MinProblem(pybnb.Problem):
    def sense(self): return pybnb.minimize

class MaxProblem(pybnb.Problem):
    def sense(self): return pybnb.maximize

@pytest.mark.skipif(not pyomo_available,
                    reason="Pyomo is not available")
class Test(object):

    def test_correct_integer_lb(self):
        # it is important that we use floating point numbers
        # that are exactly representable in base 2 so the
        # edge cases behave as expected.
        eps = 0.03125
        tol = 0.25
        for b in [-4, -3, -2, -1, 0, 1, 2, 3, 4]:
            assert correct_integer_lb(b-tol-eps,
                                      integer_tolerance=tol) \
                == b
            assert correct_integer_lb(b-tol,
                                      integer_tolerance=tol) \
                == b
            assert correct_integer_lb(b-tol+eps,
                                      integer_tolerance=tol) \
                == b
            assert correct_integer_lb(b-eps,
                                      integer_tolerance=tol) \
                == b
            assert correct_integer_lb(b,
                                      integer_tolerance=tol) \
                == b
            assert correct_integer_lb(b+eps,
                                      integer_tolerance=tol) \
                == b
            assert correct_integer_lb(b+tol-eps,
                                      integer_tolerance=tol) \
                == b
            assert correct_integer_lb(b+tol,
                                      integer_tolerance=tol) \
                == b
            assert correct_integer_lb(b+tol+eps,
                                      integer_tolerance=tol) \
                == b+1
            assert correct_integer_lb(b+1-eps,
                                      integer_tolerance=tol) \
                == b+1
            assert correct_integer_lb(b+1,
                                      integer_tolerance=tol) \
                == b+1
            assert correct_integer_lb(b+1+eps,
                                      integer_tolerance=tol) \
                == b+1
            assert correct_integer_lb(b+1+tol-eps,
                                      integer_tolerance=tol) \
                == b+1
            assert correct_integer_lb(b+1+tol,
                                      integer_tolerance=tol) \
                == b+1
            assert correct_integer_lb(b+1+tol+eps,
                                      integer_tolerance=tol) \
                == b+2

    def test_correct_integer_ub(self):
        # it is important that we use floating point numbers
        # that are exactly representable in base 2 so the
        # edge cases behave as expected.
        eps = 0.03125
        tol = 0.25
        for b in [-4, -3, -2, -1, 0, 1, 2, 3, 4]:
            assert correct_integer_ub(b-tol-eps,
                                      integer_tolerance=tol) \
                == b-1
            assert correct_integer_ub(b-tol,
                                      integer_tolerance=tol) \
                == b
            assert correct_integer_ub(b-tol+eps,
                                      integer_tolerance=tol) \
                == b
            assert correct_integer_ub(b-eps,
                                      integer_tolerance=tol) \
                == b
            assert correct_integer_ub(b,
                                      integer_tolerance=tol) \
                == b
            assert correct_integer_ub(b+eps,
                                      integer_tolerance=tol) \
                == b
            assert correct_integer_ub(b+tol-eps,
                                      integer_tolerance=tol) \
                == b
            assert correct_integer_ub(b+tol,
                                      integer_tolerance=tol) \
                == b
            assert correct_integer_ub(b+tol+eps,
                                      integer_tolerance=tol) \
                == b
            assert correct_integer_ub(b+1-eps,
                                      integer_tolerance=tol) \
                == b+1
            assert correct_integer_ub(b+1,
                                      integer_tolerance=tol) \
                == b+1
            assert correct_integer_ub(b+1+eps,
                                      integer_tolerance=tol) \
                == b+1
            assert correct_integer_ub(b+1+tol-eps,
                                      integer_tolerance=tol) \
                == b+1
            assert correct_integer_ub(b+1+tol,
                                      integer_tolerance=tol) \
                == b+1
            assert correct_integer_ub(b+1+tol+eps,
                                      integer_tolerance=tol) \
                == b+1

    def test_add_tmp_component(self):
        model = pmo.block()
        obj = pmo.objective()
        name = add_tmp_component(model,
                                 "objective",
                                 obj)
        assert name == "objective"
        assert getattr(model, name) is obj
        obj = pmo.objective()
        name = add_tmp_component(model,
                                 "objective",
                                 obj)
        assert name == ".objective."
        assert getattr(model, name) is obj

    def test_create_optimality_bound(self):
        # max
        problem = MaxProblem()
        model = pmo.block()
        model.objective = pmo.objective(sense=pmo.maximize)
        con = create_optimality_bound(problem,
                                      model.objective,
                                      100)
        assert con.ctype is pmo.constraint._ctype
        assert con.body is model.objective
        assert con.lb == 100
        # min
        problem = MinProblem()
        model = pmo.block()
        model.objective = pmo.objective(sense=pmo.minimize)
        con = create_optimality_bound(problem,
                                      model.objective,
                                      100)
        assert con.ctype is pmo.constraint._ctype
        assert con.body is model.objective
        assert con.ub == 100

    def test_sense(self):
        class Max(PyomoProblem):
            def __init__(self):
                self._pyomo_model = pmo.block()
                self._pyomo_model.o = pmo.objective(
                    sense=pmo.maximize)
                self._pyomo_model_objective = self._pyomo_model.o
                super(Max, self).__init__()
            @property
            def pyomo_model(self):
                return self._pyomo_model
            @property
            def pyomo_model_objective(self):
                return self._pyomo_model_objective
        assert Max().sense() == pybnb.maximize
        class Min(PyomoProblem):
            def __init__(self):
                self._pyomo_model = pmo.block()
                self._pyomo_model.o = pmo.objective(
                    sense=pmo.minimize)
                self._pyomo_model_objective = self._pyomo_model.o
                super(Min, self).__init__()
            @property
            def pyomo_model(self):
                return self._pyomo_model
            @property
            def pyomo_model_objective(self):
                return self._pyomo_model_objective
        assert Min().sense() == pybnb.minimize

    def test_RangeReductionProblem(self):
        class Junk(PyomoProblem):
            def __init__(self):
                self._pyomo_model = pmo.block()
                self._pyomo_model.x = pmo.variable()
                self._pyomo_model.c = pmo.constraint()
                self._pyomo_model.o = pmo.objective()
                self._pyomo_model_objective = self._pyomo_model.o
                super(Junk, self).__init__()
            @property
            def pyomo_model(self):
                return self._pyomo_model
            @property
            def pyomo_model_objective(self):
                return self._pyomo_model_objective
        junk = Junk()
        assert junk.pyomo_model_objective is junk.pyomo_model.o
        assert junk.pyomo_object_to_cid[junk.pyomo_model] == ()
        assert junk.pyomo_object_to_cid[junk.pyomo_model.x] == ('x',)
        assert junk.pyomo_object_to_cid[junk.pyomo_model.c] == ('c',)
        assert junk.pyomo_object_to_cid[junk.pyomo_model.o] == ('o',)
        assert junk.cid_to_pyomo_object[()] is junk.pyomo_model
        assert junk.cid_to_pyomo_object[('x',)] is junk.pyomo_model.x
        assert junk.cid_to_pyomo_object[('c',)] is junk.pyomo_model.c
        assert junk.cid_to_pyomo_object[('o',)] is junk.pyomo_model.o
        junk.pyomo_model.r = pmo.constraint()
        junk.update_pyomo_object_cids()
        assert junk.pyomo_object_to_cid[junk.pyomo_model.r] == ('r',)
        assert junk.cid_to_pyomo_object[('r',)] is junk.pyomo_model.r

        rr_junk = RangeReductionProblem(junk)
        assert rr_junk._best_objective == pybnb.inf
        node_ = pybnb.Node()
        node_.objective = 1
        rr_junk.notify_new_best_node(node_, False)
        assert rr_junk._best_objective == 1
        node_.objective = 2
        rr_junk.notify_new_best_node(node_, False)
        assert rr_junk._best_objective == 2



"""
Convergence checking implementation.

Copyright by Gabriel A. Hackebeil (gabe.hackebeil@gmail.com).
"""
from __future__ import division

import math

from pybnb.common import (minimize,
                          maximize,
                          inf,
                          TerminationCondition)

def compute_absolute_gap(sense, bound, objective):
    """Returns the absolute gap between the bound and
    the objective, respecting the sign relative to the
    objective sense of this problem."""
    if bound == objective:
        return 0.0
    elif math.isinf(bound) or math.isinf(objective):
        if sense == minimize:
            if (bound == -inf) or \
               (objective == inf):
                return inf
            else:
                return -inf
        else:
            if (bound == inf) or \
               (objective == -inf):
                return inf
            else:
                return -inf
    else:
        if sense == minimize:
            gap = objective - bound
        else:
            gap = bound - objective
        return gap

def _scale_absolute_gap(gap, scale):
    """Convert an absolute gap to a relative gap with the
    given scaling factor."""
    assert scale > 0
    if not math.isinf(gap):
        return gap / scale
    else:
        return gap

def _default_scale(bound, objective):
    """`max{1.0,|objective|}`"""
    # avoid using abs() as it uses the same logic, but adds
    # additional function call overhead
    if objective > 1.0:
        return objective
    elif objective < -1.0:
        return -objective
    else:
        return 1.0

def compute_relative_gap(sense,
                         bound,
                         objective,
                         scale=_default_scale):
    """Returns the relative gap between the bound and
    the objective, respecting the sign relative to the
    objective sense of this problem."""
    return _scale_absolute_gap(compute_absolute_gap(sense,
                                                    bound,
                                                    objective),
                               scale(bound, objective))

class ConvergenceChecker(object):
    """A class used to check convergence.

    Parameters
    ----------
    sense : {:obj:`minimize <pybnb.common.minimize>`, :obj:`maximize <pybnb.common.maximize>`}
        The objective sense for the problem.
    absolute_gap : float, optional
        The absolute difference between the objective and
        bound that determines optimality. (default: 1e-8)
    relative_gap : float, optional
        The relative difference between the objective and
        bound that determines optimality. (default: 1e-4)
    scale_function : function, optional
        A function with signature `f(bound, objective) ->
        float` that returns a positive scale factor used to
        convert the absolute difference between the bound
        and objective into a relative difference. The
        relative difference is compared with the
        `relative_gap` convergence tolerance to determine if
        the solver should terminate. The default is
        equivalent to `max{1.0,|objective|}`. Other examples
        one could use are `max{|bound|,|objective|}`,
        `(|bound|+|objective|)/2`, etc.
    queue_tolerance : float, optional
        The absolute tolerance used when deciding if a node
        is eligible to enter the queue. The difference
        between the node bound and the incumbent objective
        must be greater than this value. The default setting
        of zero means that nodes whose bound is equal to the
        incumbent objective are not eligible to enter the
        queue. Setting this to larger values can be used to
        control the queue size, but it should be kept small
        enough to allow absolute and relative optimality
        tolerances to be met. This option can also be set to
        `None` to allow nodes with a bound equal to (but not
        greater than) the incumbent objective to enter the
        queue. (default: 0)
    branch_tolerance : float, optional
        The absolute tolerance used when deciding if the
        computed objective and bound for a node are
        sufficiently different to branch into the node. The
        default value of zero means that branching will
        occur if the bound is not exactly equal to the
        objective. This option can be set to `None` to
        enable branching for nodes with a bound and
        objective that are exactly equal. (default: 0)
    comparison_tolerance : float, optional
        The absolute tolerance used when deciding if two
        objective or bound values are sufficiently different
        to be considered improved or worsened. This
        tolerance controls when the solver considers a new
        incumbent objective to be found. It also controls
        when warnings are output about bounds becoming worse
        on child nodes. Setting this to larger values can be
        used to avoid the above solver actions due to
        insignificant numerical differences, but it is
        better to deal with these numerical issues by
        rounding numbers to a reliable precision before
        returning them from the problem methods.
        (default: 0)
    objective_stop : float, optional
        If provided, the "objective_limit" termination
        criteria is met when a feasible objective is found
        that is at least as good as the specified value. If
        this value is infinite, the termination criteria is
        met as soon as a finite objective is found.
        (default: None)
    bound_stop : float, optional
        If provided, the "objective_limit" termination
        criteria is met when the best bound on the objective
        is at least as good as the specified value. If this
        value is infinite, the termination criteria is met
        as soon as a finite objective is found.
        (default: None)
    """
    __slots__ = ("sense",
                 "absolute_gap",
                 "relative_gap",
                 "scale_function",
                 "queue_tolerance",
                 "branch_tolerance",
                 "comparison_tolerance",
                 "objective_stop",
                 "bound_stop",
                 "infeasible_objective",
                 "unbounded_objective")

    def __init__(self,
                 sense,
                 absolute_gap=1e-8,
                 relative_gap=1e-4,
                 scale_function=_default_scale,
                 queue_tolerance=0,
                 branch_tolerance=0,
                 comparison_tolerance=0,
                 objective_stop=None,
                 bound_stop=None):
        self.sense = sense
        if self.sense == minimize:
            self.infeasible_objective = inf
            self.unbounded_objective = -inf
        else:
            assert self.sense == maximize
            self.infeasible_objective = -inf
            self.unbounded_objective = inf
        self.absolute_gap = None
        if absolute_gap is not None:
            self.absolute_gap = float(absolute_gap)
            assert (self.absolute_gap >= 0) and \
                (not math.isinf(self.absolute_gap))
        self.relative_gap = None
        if relative_gap is not None:
            self.relative_gap = float(relative_gap)
            assert self.relative_gap >= 0 and \
                (not math.isinf(self.relative_gap))
        self.scale_function = scale_function
        self.queue_tolerance = float(queue_tolerance) \
            if (queue_tolerance is not None) else queue_tolerance
        assert (self.queue_tolerance is None) or \
            ((self.queue_tolerance >= 0) and \
            (not math.isinf(self.queue_tolerance)) and \
            (not math.isnan(self.queue_tolerance)))
        self.branch_tolerance = float(branch_tolerance) \
            if (branch_tolerance is not None) else branch_tolerance
        assert (self.branch_tolerance is None) or \
            ((self.branch_tolerance >= 0) and \
            (not math.isinf(self.branch_tolerance)) and \
            (not math.isnan(self.branch_tolerance)))
        self.comparison_tolerance = float(comparison_tolerance)
        assert self.comparison_tolerance >= 0 and \
            (not math.isinf(self.comparison_tolerance)) and \
            (not math.isnan(self.comparison_tolerance))
        self.objective_stop = None
        if objective_stop is not None:
            self.objective_stop = float(objective_stop)
            assert self.objective_stop != self.unbounded_objective
            assert not math.isnan(self.objective_stop)
        self.bound_stop = None
        if bound_stop is not None:
            self.bound_stop = float(bound_stop)
            assert self.bound_stop != self.infeasible_objective
            assert not math.isnan(self.bound_stop)

    def check_termination_criteria(self,
                                   global_bound,
                                   best_objective):
        """Checks if any termination criteria are met and returns
        the corresponding :class:`TerminationCondition
        <pybnb.common.TerminationCondition>` enum value;
        otherwise, `None` is returned."""
        result = None
        if (global_bound == self.infeasible_objective) or \
           (self.objective_is_optimal(best_objective, global_bound)):
            result = TerminationCondition.optimality
        elif self.objective_stop is not None:
            if self.objective_stop != self.infeasible_objective:
                if self.sense == minimize:
                    if best_objective <= self.objective_stop:
                        result = TerminationCondition.objective_limit
                else:
                    if best_objective >= self.objective_stop:
                        result = TerminationCondition.objective_limit
            else:
                if best_objective != self.infeasible_objective:
                    result = TerminationCondition.objective_limit
        elif self.bound_stop is not None:
            if self.bound_stop != self.unbounded_objective:
                if self.sense == minimize:
                    if global_bound >= self.bound_stop:
                        result = TerminationCondition.objective_limit
                else:
                    if global_bound <= self.bound_stop:
                        result = TerminationCondition.objective_limit
            else:
                if global_bound != self.unbounded_objective:
                    result = TerminationCondition.objective_limit
        return result

    def objective_is_optimal(self, objective, bound):
        """Determines if the objective is optimal by
        checking if the optimality gap is small enough
        relative to the absolute gap or relative gap
        settings."""
        assert bound != self.infeasible_objective
        if (objective != self.unbounded_objective) and \
           (objective != self.infeasible_objective):
            gap = self.compute_absolute_gap(bound,
                                            objective)
            if (self.absolute_gap is not None) and \
               (gap <= self.absolute_gap):
                return True
            elif (self.relative_gap is not None):
                scale = self.scale_function(bound, objective)
                gap = _scale_absolute_gap(gap, scale)
                if gap <= self.relative_gap:
                    return True
        return False

    def compute_absolute_gap(self, bound, objective):
        """Returns the absolute gap between the bound and
        the objective, respecting the sign relative to the
        objective sense of this problem."""
        return compute_absolute_gap(self.sense,
                                    bound,
                                    objective)

    def compute_relative_gap(self, bound, objective):
        """Returns the relative gap between the bound and
        the objective, respecting the sign relative to the
        objective sense of this problem."""
        return compute_relative_gap(self.sense,
                                    bound,
                                    objective,
                                    scale=self.scale_function)

    def _check_eligible(self, bound, objective, tolerance):
        if (bound == self.infeasible_objective) or \
           (objective == self.unbounded_objective):
            return False
        if self.sense == minimize:
            delta = objective - bound
        else:
            delta = bound - objective
        assert not math.isnan(delta)
        if tolerance is not None:
            return delta > tolerance
        else:
            return delta >= 0

    def eligible_for_queue(self, bound, objective):
        """Returns True when the queue object with the given
        bound is eligible for the queue relative to the
        given objective."""
        return self._check_eligible(bound,
                                    objective,
                                    self.queue_tolerance)

    def eligible_to_branch(self, bound, objective):
        """Returns True when the bound and objective
        are sufficiently far apart to allow branching."""
        return self._check_eligible(bound,
                                    objective,
                                    self.branch_tolerance)

    def _check_delta(self, new, old):
        # handles the both equal and infinite case
        if old == new:
            return False
        if self.sense == minimize:
            delta = old - new
        else:
            delta = new - old
        assert not math.isnan(delta)
        return delta > self.comparison_tolerance

    def bound_worsened(self, new, old):
        """Returns True when the new bound is worse than the
        old bound by greater than the comparison
        tolerance."""
        return self._check_delta(new, old)

    def objective_improved(self, new, old):
        """Returns True when the new objective is better
        than the old objective by greater than the
        comparison tolerance."""
        return self._check_delta(new, old)

    def worst_bound(self, *args, **kwds):
        """Returns the worst bound, as defined by the
        objective sense, from a given iterable of bound
        values. This function passes all keywords and
        arguments directly to the built-ins `min` and
        `max`."""
        if self.sense == minimize:
            return min(*args)
        else:
            return max(*args)

    def best_bound(self, *args, **kwds):
        """Returns the best bound, as defined by the
        objective sense, from a given iterable of bound
        values. This function passes all keywords and
        arguments directly to the built-ins `min` and
        `max`."""
        if self.sense == minimize:
            return max(*args, **kwds)
        else:
            return min(*args, **kwds)

    def worst_objective(self, *args, **kwds):
        """Returns the worst objective, as defined by the
        objective sense, from a given iterable of objective
        values. This function passes all keywords and
        arguments directly to the built-ins `min` and
        `max`."""
        if self.sense == minimize:
            return max(*args, **kwds)
        else:
            return min(*args, **kwds)

    def best_objective(self, *args, **kwds):
        """Returns the best objective, as defined by the
        objective sense, from a given iterable of objective
        values. This function passes all keywords and
        arguments directly to the built-ins `min` and
        `max`."""
        if self.sense == minimize:
            return min(*args, **kwds)
        else:
            return max(*args, **kwds)



import pytest

from pybnb.configuration import config
from pybnb.node import (loads,
                        dumps,
                        _SerializedNode,
                        Node)

class TestNode(object):

    def test_bad_serializer(self):
        orig = config.SERIALIZER
        config.SERIALIZER = "_not_dill_or_pickle_"
        try:
            with pytest.raises(ValueError):
                loads(b'')
            with pytest.raises(ValueError):
                dumps(None)
        finally:
            config.SERIALIZER = orig
        print(config)

    def test_resize(self):
        node = Node()
        with pytest.raises(NotImplementedError):
            node.resize(1)

    def test_init(self):
        node = Node()
        assert node.objective is None
        assert node.bound is None
        assert node.tree_depth is None
        assert node.queue_priority is None
        assert node.state is None

    def test_new_child(self):
        node = Node()
        assert node.objective is None
        assert node.bound is None
        assert node.tree_depth is None
        assert node.queue_priority is None
        assert node.state is None
        node.tree_depth = 0
        node = node.new_child()
        assert node.objective is None
        assert node.bound is None
        assert node.tree_depth == 1
        assert node.queue_priority is None
        assert node.state is None
        node.objective = 1
        node.bound = -1
        node.queue_priority = 5
        node.state = 'a'
        node = node.new_child()
        assert node.objective == 1
        assert node.bound == -1
        assert node.tree_depth == 2
        assert node.queue_priority is None
        assert node.state is None

    def test_str(self):
        node = Node()
        node.objective = -1
        node.bound = -2
        node.tree_depth = 3
        node.queue_priority = (1,2,3)
        node.state = 'a'
        assert str(node) == \
            """\
Node(objective=-1,
     bound=-2,
     tree_depth=3)"""

    def test_serialization(self):
        node = Node()
        node.objective = 0.0
        node.bound = 1.0
        node.tree_depth = -1
        node.queue_priority = (1,2,3)
        node._uuid = None
        node.state = 'a'
        s = _SerializedNode.from_node(node)
        assert s.objective == node.objective
        assert s.bound == node.bound
        assert s.tree_depth == node.tree_depth
        assert s.queue_priority == node.queue_priority
        assert s._uuid is None
        assert s.data is not None
        s._generate_uuid()
        assert s._uuid is not None
        node_ = s.restore_node(s.slots)
        assert node_.objective == node.objective
        assert node_.bound == node.bound
        assert node_.tree_depth == node.tree_depth
        assert node_.queue_priority == node.queue_priority
        assert node_.state == node.state
        assert node_._uuid == s._uuid



__all__ = ('__title__',
           '__summary__',
           '__uri__',
           '__version__',
           '__author__',
           '__email__',
           '__license__',
           '__copyright__')

__title__ = 'pybnb'
__summary__ = 'A parallel branch-and-bound engine for Python'
__uri__ = 'https://github.com/ghackebeil/pybnb'
__version__ = '0.5.0'
__author__ = 'Gabriel A. Hackebeil'
__email__ = 'gabe.hackebeil@gmail.com'
__license__ = 'MIT'
__copyright__ = 'Copyright {0}'.format(__author__)



import os
import tempfile
import logging
import signal

import pytest

from pybnb.misc import (MPI_InterruptHandler,
                        metric_format,
                        time_format,
                        get_gap_labels,
                        as_stream,
                        get_default_args,
                        get_keyword_docs,
                        get_simple_logger)

from six import StringIO

yaml_available = False
try:
    import yaml
    yaml_available = True
except ImportError:
    pass

class Test(object):

    def test_MPI_InterruptHandler(self):
        assert len(MPI_InterruptHandler._sigs) > 0
        original_handlers = \
            [(signum, signal.getsignal(signum))
             for signum in MPI_InterruptHandler._sigs]
        with MPI_InterruptHandler(lambda s,f: None) as h:
            assert not h._released
        assert h._released
        for i, signum in enumerate(
                MPI_InterruptHandler._sigs):
            orig = signal.getsignal(signum)
            assert original_handlers[i][0] == signum
            assert original_handlers[i][1] is orig

        def fn(s,f):
            fn.called = True
        fn.called = False
        with MPI_InterruptHandler(fn) as h:
            assert not h._released
            signum = MPI_InterruptHandler._sigs[0]
            signal.getsignal(signum)(None, None)
            assert h._released
            for i, signum in enumerate(
                MPI_InterruptHandler._sigs):
                orig = signal.getsignal(signum)
                assert original_handlers[i][0] == signum
                assert original_handlers[i][1] is orig
        assert fn.called

    def test_metric_format(self):
        assert metric_format(None) == "<unknown>"
        assert metric_format(0.0) == "0.0 s"
        assert metric_format(0.0, align_unit=True) == "0.0 s "
        assert metric_format(0.0, unit='B') == "0.0 B"
        assert metric_format(0.0, digits=2) == "0.00 s"
        assert metric_format(1000.23, digits=3) == "1.000 Ks"
        assert metric_format(1000.23, digits=4) == "1.0002 Ks"
        assert metric_format(1000000.23, digits=4) == "1.0000 Ms"
        assert metric_format(0.23334, digits=1) == "233.3 ms"
        assert metric_format(0.23334, digits=2) == "233.34 ms"
        assert metric_format(0.00023334, digits=1) == "233.3 us"
        assert metric_format(0.00023334, digits=2) == "233.34 us"

    def test_time_format(self):
        assert time_format(None) == "<unknown>"
        assert time_format(0.0) == "0.0 s"
        assert time_format(0.0, align_unit=True) == "0.0 s "
        assert time_format(0.0, digits=2) == "0.00 s"
        assert time_format(24.9) == "24.9 s"
        assert time_format(93.462, digits=3) == "1.558 m"
        assert time_format(93.462, digits=4) == "1.5577 m"
        assert time_format(93.462, digits=4,
                           align_unit=True) == "1.5577 m "
        assert time_format(5607.72, digits=3) == "1.558 h"
        assert time_format(5607.72, digits=4) == "1.5577 h"
        assert time_format(5607.72, digits=4,
                           align_unit=True) == "1.5577 h "
        assert time_format(134585.28, digits=3) == "1.558 d"
        assert time_format(134585.28, digits=4) == "1.5577 d"
        assert time_format(134585.28, digits=4,
                           align_unit=True) == "1.5577 d "
        assert time_format(0.23334, digits=1) == "233.3 ms"
        assert time_format(0.23334, digits=2) == "233.34 ms"

    def test_get_gap_labels(self):
        l0, l1, l2 = get_gap_labels(1)
        assert l0 == 10
        assert l1 == "{gap:>10}"
        assert l2 == "{gap:>10.1f}"
        l0, l1, l2 = get_gap_labels(0.1)
        assert l0 == 10
        assert l1 == "{gap:>10}"
        assert l2 == "{gap:>10.2f}"
        l0, l1, l2 = get_gap_labels(0.01)
        assert l0 == 10
        assert l1 == "{gap:>10}"
        assert l2 == "{gap:>10.3f}"
        l0, l1, l2 = get_gap_labels(0.001)
        assert l0 == 10
        assert l1 == "{gap:>10}"
        assert l2 == "{gap:>10.4f}"
        l0, l1, l2 = get_gap_labels(0.0001)
        assert l0 == 10
        assert l1 == "{gap:>10}"
        assert l2 == "{gap:>10.5f}"
        l0, l1, l2 = get_gap_labels(0.00001)
        assert l0 == 11
        assert l1 == "{gap:>11}"
        assert l2 == "{gap:>11.6f}"
        l0, l1, l2 = get_gap_labels(0.000001,key='rgap')
        assert l0 == 12
        assert l1 == "{rgap:>12}"
        assert l2 == "{rgap:>12.7f}"
        l0, l1, l2 = get_gap_labels(0.0000001,key='agap',format='g')
        assert l0 == 13
        assert l1 == "{agap:>13}"
        assert l2 == "{agap:>13.8g}"

    def test_as_stream(self):
        fid, fname = tempfile.mkstemp()
        os.close(fid)
        with as_stream(fname) as f:
            assert not f.closed
            assert hasattr(f,'write')
        assert f.closed
        fid, fname = tempfile.mkstemp()
        os.close(fid)
        with as_stream(u""+fname) as f:
            assert not f.closed
            assert hasattr(f,'write')
        assert f.closed
        with open(fname) as f:
            assert not f.closed
            with as_stream(f) as f_:
                assert f is f_
                assert not f.closed
            assert not f.closed

    def test_get_default_args(self):
        def f(a):                                 #pragma:nocover
            pass
        assert get_default_args(f) == {}
        def f(a, b):                              #pragma:nocover
            pass
        assert get_default_args(f) == {}
        def f(*args):                             #pragma:nocover
            pass
        assert get_default_args(f) == {}
        def f(**kwds):                            #pragma:nocover
            pass
        assert get_default_args(f) == {}
        def f(*args, **kwds):                     #pragma:nocover
            pass
        assert get_default_args(f) == {}
        def f(a, b=1):                            #pragma:nocover
            pass
        assert get_default_args(f) == {'b':1}
        def f(a=1):                               #pragma:nocover
            pass
        assert get_default_args(f) == {'a':1}
        def f(a=(1,)):                            #pragma:nocover
            pass
        assert get_default_args(f) == {'a':(1,)}

    def test_get_keyword_docs(self):
        if not yaml_available:
            pytest.skip("yaml is not available")
        import pybnb.solver
        data = get_keyword_docs(pybnb.solver.Solver.solve.__doc__)
        kwds = get_default_args(pybnb.solver.Solver.solve)
        assert len(data) > 1
        for key in data:
            if 'default' in data[key]:
                assert data[key]['default'] == kwds[key]
            assert "choices" not in data[key]
        def f():
            """Something

            Parameters
            ----------
            junk1 : {"a", "b", 1}
                Junk1 description.
            junk2 : {"c", "d"}, optional
                Junk2 description more than one
                line. (default: "c")
            junk3 : int
                Junk3 description.
            """
        data = get_keyword_docs(f.__doc__)
        assert data == \
            {'junk1': {'choices': ['a', 'b', 1],
                       'doc': 'Junk1 description.'},
             'junk2': {'choices': ['c', 'd'],
                       'default': 'c',
                       'doc': 'Junk2 description more than one line.'},
             'junk3': {'doc': 'Junk3 description.'}}

    def test_get_simple_logger(self):
        log = get_simple_logger(console=False)
        assert log.disabled
        log = get_simple_logger()
        assert not log.disabled
        log = get_simple_logger(console=True)
        assert not log.disabled
        assert len(log.handlers) == 2
        log.info('junk')
        fid, fname = tempfile.mkstemp()
        out = StringIO()
        os.close(fid)
        formatter = logging.Formatter("[%(levelname)s] %(message)s")
        try:
            log = get_simple_logger(filename=fname,
                                    stream=out,
                                    console=True,
                                    formatter=formatter,
                                    level=logging.WARNING)
            assert len(log.handlers) == 4
            log.error('error_line')
            log.warning('warning_line')
            log.info('info_line')
            log.debug('debug_line')
            for handler in log.handlers:
                handler.close()
            with open(fname) as f:
                lines = f.readlines()
                assert len(lines) == 2
                assert lines[0].strip() == '[ERROR] error_line'
                assert lines[1].strip() == '[WARNING] warning_line'
                del lines
            lines = out.getvalue().splitlines()
            assert lines[0].strip() == '[ERROR] error_line'
            assert lines[1].strip() == '[WARNING] warning_line'
        finally:
            os.remove(fname)



import os
import sys
import setuptools
import setuptools.command.test
from codecs import open

here = os.path.abspath(os.path.dirname(__file__))

about = {}
with open(os.path.join(here, "src", "pybnb", "__about__.py")) as f:
    exec(f.read(), about)

# Get the long description from the README file
def _readme():
    with open(os.path.join(here, "README.rst"), encoding="utf-8") as f:
        return f.read()

install_requires = []
with open(os.path.join(here, "requirements.txt")) as f:
    install_requires.extend([ln.strip() for ln in f
                             if ln.strip() != ""])
if sys.version_info < (3, 4):
    # Backport of Python 3.4 enums to earlier versions
    install_requires.append('enum34')

tests_require = []
with open(os.path.join(here, "test_requirements.txt")) as f:
    tests_require.extend([ln.strip() for ln in f
                          if (ln.strip() != "") and \
                          (not ln.strip() == "-r requirements.txt")])

class PyTest(setuptools.command.test.test):
    user_options = [('pytest-args=', 'a', "Arguments to pass to pytest")]

    def initialize_options(self):
        super(PyTest, self).initialize_options()
        self.pytest_args = ''

    def run_tests(self):
        import shlex
        #import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(shlex.split(self.pytest_args))
        sys.exit(errno)

setuptools.setup(
    name=about["__title__"],
    version=about["__version__"],
    description=about["__summary__"],
    long_description=_readme(),
    url=about["__uri__"],
    author=about["__author__"],
    author_email=about["__email__"],
    license=about["__license__"],
    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Natural Language :: English",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    keywords=["optimization","branch and bound"],
    packages=setuptools.find_packages('src', exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    package_dir={'':'src'},
    install_requires=install_requires,
    cmdclass={
        "test": PyTest
    },
    # use MANIFEST.in
    include_package_data=True,
    tests_require=tests_require
)



import itertools
import math

from pybnb.common import (minimize,
                          maximize,
                          inf,
                          TerminationCondition)
from pybnb.convergence_checker import ConvergenceChecker

class TestConvergenceChecker(object):

    def test_sense(self):
        # min
        p = ConvergenceChecker(minimize)
        assert p.sense == minimize
        # max
        p = ConvergenceChecker(maximize)
        assert p.sense == maximize

    def test_infeasible_objective(self):
        # min
        p = ConvergenceChecker(minimize)
        assert p.infeasible_objective == inf
        # max
        p = ConvergenceChecker(maximize)
        assert p.infeasible_objective == -inf

    def test_unbounded_objective(self):
        # min
        p = ConvergenceChecker(minimize)
        assert p.unbounded_objective == -inf
        # max
        p = ConvergenceChecker(maximize)
        assert p.unbounded_objective == inf

    def test_check_termination_criteria(self):
        # min
        p = ConvergenceChecker(minimize)
        assert p.check_termination_criteria(inf, None) is \
            TerminationCondition.optimality
        assert p.check_termination_criteria(0, 0) is \
            TerminationCondition.optimality
        assert p.check_termination_criteria(0, 1) is None
        p = ConvergenceChecker(minimize,
                               objective_stop=1)
        assert p.check_termination_criteria(0, 1.1) is None
        assert p.check_termination_criteria(0, 1) is \
            TerminationCondition.objective_limit
        p = ConvergenceChecker(minimize,
                               objective_stop=inf)
        assert p.check_termination_criteria(-inf, inf) is None
        assert p.check_termination_criteria(-inf, 1000000) is \
            TerminationCondition.objective_limit
        p = ConvergenceChecker(minimize,
                               bound_stop=0)
        assert p.check_termination_criteria(-0.1, 1) is None
        assert p.check_termination_criteria(0, 1) is \
            TerminationCondition.objective_limit
        p = ConvergenceChecker(minimize,
                               bound_stop=-inf)
        assert p.check_termination_criteria(-inf, inf) is None
        assert p.check_termination_criteria(-10000000, inf) is \
            TerminationCondition.objective_limit
        # max
        p = ConvergenceChecker(maximize)
        assert p.check_termination_criteria(-inf, None) is \
            TerminationCondition.optimality
        assert p.check_termination_criteria(0, 0) is \
            TerminationCondition.optimality
        assert p.check_termination_criteria(0, -1) is None
        p = ConvergenceChecker(maximize,
                               objective_stop=-1)
        assert p.check_termination_criteria(0, -1.1) is None
        assert p.check_termination_criteria(0, -1) is \
            TerminationCondition.objective_limit
        p = ConvergenceChecker(maximize,
                               objective_stop=-inf)
        assert p.check_termination_criteria(inf, -inf) is None
        assert p.check_termination_criteria(inf, -1000000) is \
            TerminationCondition.objective_limit
        p = ConvergenceChecker(maximize,
                               bound_stop=0)
        assert p.check_termination_criteria(0.1, -1) is None
        assert p.check_termination_criteria(0, -1) is \
            TerminationCondition.objective_limit
        p = ConvergenceChecker(maximize,
                               bound_stop=inf)
        assert p.check_termination_criteria(inf, -inf) is None
        assert p.check_termination_criteria(1000000, -inf) is \
            TerminationCondition.objective_limit

    def test_objective_is_optimal(self):

        for sense in [minimize, maximize]:
            p = ConvergenceChecker(sense)
            for bound,objective in itertools.product([-inf,
                                                      inf,
                                                      0.0],
                                                     [-inf,
                                                      inf,
                                                      0.0]):
                if math.isinf(bound) and math.isinf(objective):
                    if bound != p.infeasible_objective:
                        assert not p.objective_is_optimal(objective, bound)
                elif objective == bound:
                    assert p.objective_is_optimal(objective, bound)
                elif bound != p.infeasible_objective:
                    assert not p.objective_is_optimal(objective, bound)
            p = ConvergenceChecker(sense,
                                   absolute_gap=0,
                                   relative_gap=0)
            assert p.objective_is_optimal(0,0)
            assert p.objective_is_optimal(1,1)
            assert p.objective_is_optimal(-1,-1)

    def test_compute_relative_gap(self):

        for sense in [minimize, maximize]:
            p = ConvergenceChecker(sense)
            for bound,objective in itertools.product([-inf,
                                                      inf,
                                                      0.0],
                                                     [-inf,
                                                      inf,
                                                      0.0]):
                if (not math.isinf(bound)) or \
                   (not math.isinf(objective)):
                    continue
                if bound == objective:
                    assert p.compute_relative_gap(bound,objective) == 0
                elif p.sense == minimize:
                    if (bound == -inf) or \
                       (objective == inf):
                        assert p.compute_relative_gap(bound,objective) == \
                            inf
                    else:
                        assert p.compute_relative_gap(bound,objective) == \
                            -inf
                else:
                    assert p.sense == maximize
                    if (bound == inf) or \
                       (objective == -inf):
                        assert p.compute_relative_gap(bound,objective) == \
                            inf
                    else:
                        assert p.compute_relative_gap(bound,objective) == \
                            -inf

    def test_eligible_for_queue(self):
        # min
        p = ConvergenceChecker(minimize)
        assert not p.eligible_for_queue(-inf, -inf)
        assert p.eligible_for_queue(-inf, inf)
        assert p.eligible_for_queue(-inf, 0.0)
        assert not p.eligible_for_queue(inf, -inf)
        assert not p.eligible_for_queue(inf, inf)
        assert not p.eligible_for_queue(inf, 0.0)
        assert not p.eligible_for_queue(0.0, -inf)
        assert p.eligible_for_queue(0.0, inf)
        assert p.eligible_for_queue(-1e-16, 0.0)
        assert not p.eligible_for_queue(0.0, 0.0)
        p = ConvergenceChecker(minimize,
                               queue_tolerance=0.1)
        assert not p.eligible_for_queue(-0.1, 0.0)
        assert p.eligible_for_queue(-0.11, 0.0)
        p = ConvergenceChecker(minimize,
                               queue_tolerance=None)
        assert p.eligible_for_queue(0.0, 0.0)
        # max
        p = ConvergenceChecker(maximize)
        assert not p.eligible_for_queue(-inf, -inf)
        assert not p.eligible_for_queue(-inf, inf)
        assert not p.eligible_for_queue(-inf, 0.0)
        assert p.eligible_for_queue(inf, -inf)
        assert not p.eligible_for_queue(inf, inf)
        assert p.eligible_for_queue(inf, 0.0)
        assert p.eligible_for_queue(0.0, -inf)
        assert not p.eligible_for_queue(0.0, inf)
        assert p.eligible_for_queue(1e-16, 0.0)
        assert not p.eligible_for_queue(0.0, 0.0)
        p = ConvergenceChecker(maximize,
                               queue_tolerance=0.1)
        assert not p.eligible_for_queue(0.1, 0.0)
        assert p.eligible_for_queue(0.11, 0.0)
        p = ConvergenceChecker(maximize,
                               queue_tolerance=None)
        assert p.eligible_for_queue(0.0, 0.0)

    def test_eligible_to_branch(self):
        # min
        p = ConvergenceChecker(minimize)
        assert not p.eligible_to_branch(-inf, -inf)
        assert p.eligible_to_branch(-inf, inf)
        assert p.eligible_to_branch(-inf, 0.0)
        assert not p.eligible_to_branch(inf, -inf)
        assert not p.eligible_to_branch(inf, inf)
        assert not p.eligible_to_branch(inf, 0.0)
        assert not p.eligible_to_branch(0.0, -inf)
        assert p.eligible_to_branch(0.0, inf)
        assert not p.eligible_to_branch(0.0, 0.0)
        p = ConvergenceChecker(minimize,
                               branch_tolerance=0.1)
        assert not p.eligible_to_branch(-0.1, 0.0)
        assert p.eligible_to_branch(-0.11, 0.0)
        p = ConvergenceChecker(minimize,
                               branch_tolerance=None)
        assert p.eligible_to_branch(0.0, 0.0)
        # max
        p = ConvergenceChecker(maximize)
        assert not p.eligible_to_branch(-inf, -inf)
        assert not p.eligible_to_branch(-inf, inf)
        assert not p.eligible_to_branch(-inf, 0.0)
        assert p.eligible_to_branch(inf, -inf)
        assert not p.eligible_to_branch(inf, inf)
        assert p.eligible_to_branch(inf, 0.0)
        assert p.eligible_to_branch(0.0, -inf)
        assert not p.eligible_to_branch(0.0, inf)
        assert not p.eligible_to_branch(0.0, 0.0)
        p = ConvergenceChecker(maximize,
                               branch_tolerance=0.1)
        assert not p.eligible_to_branch(0.1, 0.0)
        assert p.eligible_to_branch(0.11, 0.0)
        p = ConvergenceChecker(maximize,
                               branch_tolerance=None)
        assert p.eligible_to_branch(0.0, 0.0)

    def test_bound_worsened(self):
        # min
        p = ConvergenceChecker(minimize)
        assert not p.bound_worsened(-inf, -inf)
        assert p.bound_worsened(-inf, inf)
        assert p.bound_worsened(-inf, 0.0)
        assert not p.bound_worsened(inf, -inf)
        assert not p.bound_worsened(inf, inf)
        assert not p.bound_worsened(inf, 0.0)
        assert not p.bound_worsened(0.0, -inf)
        assert p.bound_worsened(0.0, inf)
        assert not p.bound_worsened(0.0, 0.0)
        p = ConvergenceChecker(minimize,
                               comparison_tolerance=0.1)
        assert not p.bound_worsened(-0.1, 0.0)
        assert p.bound_worsened(-0.11, 0.0)
        # max
        p = ConvergenceChecker(maximize)
        assert not p.bound_worsened(-inf, -inf)
        assert not p.bound_worsened(-inf, inf)
        assert not p.bound_worsened(-inf, 0.0)
        assert p.bound_worsened(inf, -inf)
        assert not p.bound_worsened(inf, inf)
        assert p.bound_worsened(inf, 0.0)
        assert p.bound_worsened(0.0, -inf)
        assert not p.bound_worsened(0.0, inf)
        assert not p.bound_worsened(0.0, 0.0)
        p = ConvergenceChecker(maximize,
                               comparison_tolerance=0.1)
        assert not p.bound_worsened(0.1, 0.0)
        assert p.bound_worsened(0.11, 0.0)

    def test_objective_improved(self):
        # min
        p = ConvergenceChecker(minimize)
        assert not p.objective_improved(-inf, -inf)
        assert p.objective_improved(-inf, inf)
        assert p.objective_improved(-inf, 0.0)
        assert not p.objective_improved(inf, -inf)
        assert not p.objective_improved(inf, inf)
        assert not p.objective_improved(inf, 0.0)
        assert not p.objective_improved(0.0, -inf)
        assert p.objective_improved(0.0, inf)
        assert not p.objective_improved(0.0, 0.0)
        p = ConvergenceChecker(minimize,
                               comparison_tolerance=0.1)
        assert not p.objective_improved(-0.1, 0.0)
        assert p.objective_improved(-0.11, 0.0)
        # max
        p = ConvergenceChecker(maximize)
        assert not p.objective_improved(-inf, -inf)
        assert not p.objective_improved(-inf, inf)
        assert not p.objective_improved(-inf, 0.0)
        assert p.objective_improved(inf, -inf)
        assert not p.objective_improved(inf, inf)
        assert p.objective_improved(inf, 0.0)
        assert p.objective_improved(0.0, -inf)
        assert not p.objective_improved(0.0, inf)
        assert not p.objective_improved(0.0, 0.0)
        p = ConvergenceChecker(maximize,
                               comparison_tolerance=0.1)
        assert not p.objective_improved(0.1, 0.0)
        assert p.objective_improved(0.11, 0.0)

    def test_worst_bound(self):
        # min
        p = ConvergenceChecker(minimize)
        assert p.worst_bound(-1, 0, 1) == -1
        assert p.worst_bound([-1, 0, 1]) == -1
        # max
        p = ConvergenceChecker(maximize)
        assert p.worst_bound(-1, 0, 1) == 1
        assert p.worst_bound([-1, 0, 1]) == 1

    def test_best_bound(self):
        # min
        p = ConvergenceChecker(minimize)
        assert p.best_bound(-1, 0, 1) == 1
        assert p.best_bound([-1, 0, 1]) == 1
        # max
        p = ConvergenceChecker(maximize)
        assert p.best_bound(-1, 0, 1) == -1
        assert p.best_bound([-1, 0, 1]) == -1

    def test_worst_objective(self):
        # min
        p = ConvergenceChecker(minimize)
        assert p.worst_objective(-1, 0, 1) == 1
        assert p.worst_objective([-1, 0, 1]) == 1
        # max
        p = ConvergenceChecker(maximize)
        assert p.worst_objective(-1, 0, 1) == -1
        assert p.worst_objective([-1, 0, 1]) == -1

    def test_best_objective(self):
        # min
        p = ConvergenceChecker(minimize)
        assert p.best_objective(-1, 0, 1) == -1
        assert p.best_objective([-1, 0, 1]) == -1
        # max
        p = ConvergenceChecker(maximize)
        assert p.best_objective(-1, 0, 1) == 1
        assert p.best_objective([-1, 0, 1]) == 1



import math

import pytest

from pybnb.common import (inf,
                          nan,
                          SolutionStatus,
                          TerminationCondition)
from pybnb.node import (Node, loads)
from pybnb.solver_results import SolverResults

from six import StringIO

yaml_available = False
try:
    import yaml
    yaml_available = True
except ImportError:
    pass

class TestSolverResults(object):

    def test_pprint(self):
        results = SolverResults()
        out = StringIO()
        results.pprint(stream=out)
        assert out.getvalue() == \
            """solver results:
 - solution_status: None
 - termination_condition: None
 - objective: None
 - bound: None
 - absolute_gap: None
 - relative_gap: None
 - nodes: None
 - wall_time: None
 - best_node: None
"""
        results.best_node = Node()
        out = StringIO()
        results.pprint(stream=out)
        assert out.getvalue() == \
            """solver results:
 - solution_status: None
 - termination_condition: None
 - objective: None
 - bound: None
 - absolute_gap: None
 - relative_gap: None
 - nodes: None
 - wall_time: None
 - best_node: Node(objective=None)
"""
        results.solution_status = "optimal"
        results.termination_condition = "optimality"
        results.objective = 1
        results.bound = 0
        results.absolute_gap = 1
        results.relative_gap = 1
        results.nodes = 1
        results.wall_time = 60
        results.best_node = Node()
        results.best_node.tree_depth = 3
        results.best_node.objective = 4
        results.junk0 = 1.0
        results.junk1 = 1
        results.junk2 = inf
        results.junk3 = -inf
        results.junk4 = nan
        results.junk5 = 'inf'
        results.junk6 = '-inf'
        results.junk7 = 'nan'
        results.junk8 = None
        results.junk9 = 'None'
        out = StringIO()
        results.pprint(stream=out)
        assert out.getvalue() == \
            """solver results:
 - solution_status: optimal
 - termination_condition: optimality
 - objective: 1
 - bound: 0
 - absolute_gap: 1
 - relative_gap: 1
 - nodes: 1
 - wall_time: 1.00 m
 - best_node: Node(objective=4)
 - junk0: 1.0
 - junk1: 1
 - junk2: inf
 - junk3: -inf
 - junk4: nan
 - junk5: 'inf'
 - junk6: '-inf'
 - junk7: 'nan'
 - junk8: None
 - junk9: 'None'
"""
        results.solution_status = \
            SolutionStatus.optimal
        results.termination_condition = \
            TerminationCondition.optimality
        results.objective = 1
        results.bound = 0
        results.absolute_gap = 1
        results.relative_gap = 1
        results.nodes = 1
        results.wall_time = 60
        results.best_node = Node()
        results.best_node.tree_depth = 3
        results.best_node.objective = 4
        results.junk0 = 1.0
        results.junk1 = 1
        results.junk2 = inf
        results.junk3 = -inf
        results.junk4 = nan
        results.junk5 = 'inf'
        results.junk6 = '-inf'
        results.junk7 = 'nan'
        results.junk8 = None
        results.junk9 = 'None'
        out = StringIO()
        results.pprint(stream=out)
        assert out.getvalue() == \
            """solver results:
 - solution_status: optimal
 - termination_condition: optimality
 - objective: 1
 - bound: 0
 - absolute_gap: 1
 - relative_gap: 1
 - nodes: 1
 - wall_time: 1.00 m
 - best_node: Node(objective=4)
 - junk0: 1.0
 - junk1: 1
 - junk2: inf
 - junk3: -inf
 - junk4: nan
 - junk5: 'inf'
 - junk6: '-inf'
 - junk7: 'nan'
 - junk8: None
 - junk9: 'None'
"""

    def test_write(self):
        if not yaml_available:
            pytest.skip("yaml is not available")
        results = SolverResults()
        out = StringIO()
        results.write(out)
        x = yaml.load(out.getvalue())
        assert len(x) == 9
        assert x['solution_status'] is None
        assert x['termination_condition'] is None
        assert x['objective'] is None
        assert x['bound'] is None
        assert x['absolute_gap'] is None
        assert x['relative_gap'] is None
        assert x['nodes'] is None
        assert x['wall_time'] is None
        assert x['best_node'] is None
        results.solution_status = "optimal"
        results.termination_condition = "optimality"
        del results.objective
        results.bound = inf
        results.absolute_gap = -inf
        results.relative_gap = nan
        results.nodes = 1
        results.wall_time = 1.0
        results.best_node = Node()
        results.best_node.queue_priority = 12
        results.best_node.tree_depth = 3
        results.best_node.state = (1,2)
        results.best_node.objective = 1.5
        results.best_node.bound = -1.4
        results.junk0 = 1.0
        results.junk1 = 1
        results.junk2 = inf
        results.junk3 = -inf
        results.junk4 = nan
        results.junk5 = 'inf'
        results.junk6 = '-inf'
        results.junk7 = 'nan'
        out = StringIO()
        results.write(out)
        x = yaml.load(out.getvalue())
        assert len(x) == 16
        assert x['solution_status'] == "optimal"
        assert x['termination_condition'] == "optimality"
        assert x['bound'] == inf
        assert x['absolute_gap'] == -inf
        assert math.isnan(x['relative_gap'])
        assert x['nodes'] == 1
        assert type(x['nodes']) is int
        assert x['wall_time'] == 1.0
        assert type(x['wall_time']) is float
        best_node_ = loads(x['best_node'])
        assert best_node_.queue_priority == 12
        assert best_node_.tree_depth == 3
        assert best_node_.state == (1,2)
        assert best_node_.objective == 1.5
        assert best_node_.bound == -1.4
        assert x['junk0'] == 1.0
        assert type(x['junk0']) is float
        assert x['junk1'] == 1
        assert type(x['junk1']) is int
        assert x['junk2'] == inf
        assert x['junk3'] == -inf
        assert math.isnan(x['junk4'])
        assert x['junk5'] == 'inf'
        assert x['junk6'] == '-inf'
        assert x['junk7'] == 'nan'

        results.bound = 'inf'
        results.absolute_gap = '-inf'
        results.relative_gap = 'nan'
        results.nodes = None
        results.junk0 = None
        out = StringIO()
        results.write(out)
        x = yaml.load(out.getvalue())
        assert len(x) == 16
        assert x['bound'] == 'inf'
        assert x['absolute_gap'] == '-inf'
        assert x['relative_gap'] == 'nan'
        assert x['nodes'] is None
        assert x['junk0'] is None



"""
A Base class for defining a branch-and-bound problem based
on a pyomo.kernel model.

Copyright by Gabriel A. Hackebeil (gabe.hackebeil@gmail.com).
"""
from pybnb import (minimize,
                   maximize,
                   Problem)
from pybnb.pyomo.misc import generate_cids

import pyomo.kernel as pmo

class PyomoProblem(Problem):
    """An extension of the :class:`pybnb.Problem
    <pybnb.problem.Problem>` base class for defining
    problems with a core Pyomo model."""

    def __init__(self, *args, **kwds):
        super(PyomoProblem, self).__init__(*args, **kwds)
        self.__pyomo_object_to_cid = None
        self.__cid_to_pyomo_object = None
        self.update_pyomo_object_cids()

    def update_pyomo_object_cids(self):
        (self.__pyomo_object_to_cid,
         self.__cid_to_pyomo_object) = \
            generate_cids(self.pyomo_model,
                          active=None)
        assert len(set(self.pyomo_object_to_cid.values())) == \
            len(self.__pyomo_object_to_cid)
        assert len(self.cid_to_pyomo_object) == \
            len(self.__pyomo_object_to_cid)

    @property
    def pyomo_object_to_cid(self):
        """The map from pyomo model object to component id."""
        return self.__pyomo_object_to_cid

    @property
    def cid_to_pyomo_object(self):
        """The map from component id to pyomo model object."""
        return self.__cid_to_pyomo_object

    #
    # Abstract Methods
    #

    @property
    def pyomo_model(self):
        """Returns the pyomo model for this problem.

        Note
        ----
        This method is abstract and must be defined by the
        user.
        """
        raise NotImplementedError()               #pragma:nocover

    @property
    def pyomo_model_objective(self):
        """Returns the pyomo model objective for this
        problem.

        Note
        ----
        This method is abstract and must be defined by the
        user.
        """
        raise NotImplementedError()               #pragma:nocover

    #
    # Implements a few problem methods
    #

    def sense(self):
        obj = self.pyomo_model_objective
        if obj.sense == pmo.minimize:
            return minimize
        else:
            assert obj.sense == pmo.maximize
            return maximize



import sys

import pybnb

is_pypy = False
try:
    import __pypy__
    is_pypy = True
except ImportError:
    is_pypy = False

class Test(object):

    # See what Python versions the combined
    # coverage report includes
    def test_show_coverage(self):
        if not is_pypy:
            if sys.version_info.major == 2:
                if sys.version_info.minor == 7:
                    print(sys.version_info)
            elif sys.version_info.major == 3:
                if sys.version_info.minor == 5:
                    print(sys.version_info)
                elif sys.version_info.minor == 6:
                    print(sys.version_info)
                elif sys.version_info.minor == 7:
                    print(sys.version_info)
        if is_pypy:
            if sys.version_info.major == 2:
                print(sys.version_info)
            if sys.version_info.major == 3:
                print(sys.version_info)

    def test_version(self):
        pybnb.__version__



import copy
import logging
import math

from pybnb.dispatcher import DispatcherQueueData
from pybnb.problem import _ProblemWithSolveInfoCollection

class _RedirectHandler(logging.Handler):
    """Redirects log messages with a WARNING level or above
    to the provided dispatcher's log."""

    def __init__(self, dispatcher):
        super(_RedirectHandler, self).__init__()
        self._dispatcher = dispatcher

    def emit(self, record):
        if logging.WARNING <= record.levelno < logging.ERROR:
            self._dispatcher.log_warning(record.getMessage())
        elif logging.ERROR <= record.levelno < logging.CRITICAL:
            self._dispatcher.log_error(record.getMessage())
        elif logging.CRITICAL <= record.levelno:
            self._dispatcher.log_critical(record.getMessage())

class NestedSolver(_ProblemWithSolveInfoCollection):
    """A class for creating a nested branch-and-bound solve
    strategy. An instance of this class (wrapped around a
    standard problem) can be passed to the solver as the
    problem argument.

    Parameters
    ----------
    problem : :class:`pybnb.Problem <pybnb.problem.Problem>`
        An object defining a branch-and-bound problem.
    node_limit : int, optional
        The same as the standard solver option, but applied
        to the nested solver to limit the number of nodes to
        explore when processing a work item. (default: None)
    time_limit : float, optional
        The same as the standard solver option, but applied
        to the nested solver to limit the amount of time
        spent processing a work item. (default: 5)
    queue_strategy : :class:`QueueStrategy <pybnb.common.QueueStrategy>` or tuple
        The same as the standard solver option, but applied
        to the nested solver to control the queue strategy
        used when processing a work item. (default: 'depth')
    """

    def __init__(self,
                 problem,
                 node_limit=None,
                 time_limit=5,
                 queue_strategy="depth"):
        # we import from pybnb.solver here to avoid a
        # circular import reference
        from pybnb.solver import Solver
        self._problem = problem
        self._node_limit = node_limit
        self._time_limit = time_limit
        self._queue_strategy = queue_strategy
        self._solver = Solver(comm=None)
        self._log = logging.Logger(None,
                                   level=logging.WARNING)
        self._convergence_checker = None
        self._best_objective = None
        self._best_node = None
        self._current_node = None
        self._results = None
        self._queue = None
        super(NestedSolver, self).__init__()

    def _initialize(self,
                    dispatcher,
                    best_objective,
                    disable_objective_call):
        assert best_objective is not None
        self._best_objective = best_objective
        self._disable_objective_call = disable_objective_call
        self._log.addHandler(_RedirectHandler(dispatcher))

    def _solve(self):
        bound_stop = None
        if not math.isinf(self._best_objective):
            bound_stop = self._best_objective
        # shallow copy
        root = copy.copy(self._current_node)
        init_queue = DispatcherQueueData(
            nodes=[root],
            worst_terminal_bound=None,
            sense=self._convergence_checker.sense)
        self._results = self._solver.solve(
            self._problem,
            best_objective=self._best_objective,
            best_node=self._best_node,
            bound_stop=bound_stop,
            initialize_queue=init_queue,
            log=self._log,
            absolute_gap=\
                self._convergence_checker.absolute_gap,
            relative_gap=\
                self._convergence_checker.relative_gap,
            scale_function=\
                self._convergence_checker.scale_function,
            queue_tolerance=\
                self._convergence_checker.queue_tolerance,
            branch_tolerance=\
                self._convergence_checker.branch_tolerance,
            comparison_tolerance=\
                self._convergence_checker.comparison_tolerance,
            objective_stop=\
                self._convergence_checker.objective_stop,
            disable_objective_call=self._disable_objective_call,
            node_limit=self._node_limit,
            time_limit=self._time_limit,
            queue_strategy=self._queue_strategy,
            disable_signal_handlers=True)
        self._queue = self._solver.save_dispatcher_queue()
        self._solve_info.add_from(
            self._solver._global_solve_info)

    #
    # Ducktype a partial Problem interface
    #

    def objective(self):                          #pragma:nocover
        """The solver does not call this method when it sees the
        problem implements a nested solve."""
        raise NotImplementedError()

    def bound(self):                              #pragma:nocover
        """The solver does not call this method when it sees the
        problem implements a nested solve."""
        raise NotImplementedError()

    def branch(self):                #pragma:nocover
        """The solver does not call this method when it sees the
        problem implements a nested solve."""
        raise NotImplementedError()

    def sense(self):
        """Calls the sense() method on the user-provided
        problem."""
        return self._problem.sense()

    def save_state(self, node):
        """Calls the save_state() method on the user-provided
        problem."""
        self._problem.save_state(node)

    def load_state(self, node):
        """Calls the load_state() method on the user-provided
        problem and prepares for a nested solve."""
        self._problem.load_state(node)
        self._results = None
        self._children = None
        self._current_node = node
        self._solve_found_best_node = False

    def notify_solve_begins(self,
                            comm,
                            worker_comm,
                            convergence_checker):
        """Calls the notify_solve_begins() method on the
        user-provided problem and prepares for a solve."""
        self._best_objective = None
        self._best_node = None
        self._convergence_checker = convergence_checker
        self._current_node = None
        self._results = None
        self._queue = None
        self._problem.notify_solve_begins(
            comm,
            worker_comm,
            convergence_checker)

    def notify_new_best_node(self, node, current):
        """Calls the notify_new_best_node() method on the
        user-provided problem and stores the best node for
        use in the next nested solve."""
        self._best_objective = self._convergence_checker.\
            best_objective(self._best_objective,
                           node.objective)
        self._best_node = node
        self._problem.notify_new_best_node(node, current)

    def notify_solve_finished(self,
                              comm,
                              worker_comm,
                              results):
        """Calls the notify_solve_finished() method on the
        user-provided problem and does some final
        cleanup."""
        while len(self._log.handlers) > 0:
            self._log.removeHandler(self._log.handlers[0])
        self._problem.notify_solve_finished(
            comm,
            worker_comm,
            results)



"""
Various tools for implementing branch-and-bound problems
that are based on a pyomo.kernel model.

Copyright by Gabriel A. Hackebeil (gabe.hackebeil@gmail.com).
"""
import pybnb.pyomo.misc
import pybnb.pyomo.problem
import pybnb.pyomo.range_reduction



from pybnb.common import (minimize,
                          maximize,
                          inf)
from pybnb.node import Node
from pybnb.problem import Problem
from pybnb.solver import Solver
from pybnb.dispatcher import DispatcherQueueData

from .common import mpi_available

from runtests.mpi import MPITest

class DummyProblem(Problem):
    def __init__(self, sense):
        self._sense = sense
        self._notify_new_best_node_args = None
        self._notify_new_best_node_call_count = 0
    def sense(self): return self._sense
    def objective(self): return 0
    def bound(self): return 0
    def save_state(self, node): pass
    def load_state(self, node): pass
    def branch(self): return ()
    def notify_new_best_node(self, node, current):
        self._notify_new_best_node_args = (node, current)
        self._notify_new_best_node_call_count += 1

def _test_initialize_queue(comm):
    solver = Solver(comm=comm)

    # no initial queue
    for sense in (minimize, maximize):
        problem = DummyProblem(sense)
        results = solver.solve(problem)
        assert results.solution_status == "optimal"
        assert results.termination_condition == "optimality"
        assert results.objective == 0
        assert results.bound == 0
        assert results.absolute_gap == 0
        assert results.relative_gap == 0
        assert results.nodes == 1
        assert results.wall_time is not None
        assert results.best_node is not None
        assert results.best_node.objective == results.objective
        assert results.best_node.tree_depth == 0
        if solver.is_worker:
            assert problem._notify_new_best_node_call_count == 1
            assert problem._notify_new_best_node_args[0] is results.best_node
            if solver._local_solve_info.explored_nodes_count == 1:
                assert problem._notify_new_best_node_args[1]
            else:
                assert solver._local_solve_info.explored_nodes_count == 0
                assert not problem._notify_new_best_node_args[1]
        else:
            assert problem._notify_new_best_node_call_count == 0
        problem = DummyProblem(sense)
        results = solver.solve(problem,
                               best_objective=(1 if (sense == minimize) else -1))
        assert results.solution_status == "optimal"
        assert results.termination_condition == "optimality"
        assert results.objective == 0
        assert results.bound == 0
        assert results.absolute_gap == 0
        assert results.relative_gap == 0
        assert results.nodes == 1
        assert results.wall_time is not None
        assert results.best_node is not None
        assert results.best_node.objective == results.objective
        assert results.best_node.tree_depth == 0
        if solver.is_worker:
            assert problem._notify_new_best_node_call_count == 1
            assert problem._notify_new_best_node_args[0] is results.best_node
            if solver._local_solve_info.explored_nodes_count == 1:
                assert problem._notify_new_best_node_args[1]
            else:
                assert solver._local_solve_info.explored_nodes_count == 0
                assert not problem._notify_new_best_node_args[1]
        else:
            assert problem._notify_new_best_node_call_count == 0
        problem = DummyProblem(sense)
        results = solver.solve(problem,
                               best_objective=(1 if (sense == minimize) else -1),
                               disable_objective_call=True)
        assert results.solution_status == "feasible"
        assert results.termination_condition == "no_nodes"
        assert results.objective == (1 if (sense == minimize) else -1)
        assert results.bound == 0
        assert results.absolute_gap == 1
        assert results.relative_gap == 1
        assert results.nodes == 1
        assert results.wall_time is not None
        assert results.best_node is None
        assert problem._notify_new_best_node_call_count == 0
        best_node_ = Node()
        best_node_.objective = (1 if (sense == minimize) else -1)
        best_node_._uuid = 'abcd'
        problem = DummyProblem(sense)
        results = solver.solve(problem,
                               best_node=best_node_,
                               disable_objective_call=True)
        assert results.solution_status == "feasible"
        assert results.termination_condition == "no_nodes"
        assert results.objective == best_node_.objective
        assert results.bound == 0
        assert results.absolute_gap == 1
        assert results.relative_gap == 1
        assert results.nodes == 1
        assert results.best_node._uuid == best_node_._uuid
        if (comm is None) or (comm.size == 1):
            assert results.best_node is best_node_
        assert results.best_node.objective == results.objective
        if solver.is_worker:
            assert problem._notify_new_best_node_call_count == 1
            assert problem._notify_new_best_node_args[0] is results.best_node
            if solver._local_solve_info.explored_nodes_count == 1:
                assert not problem._notify_new_best_node_args[1]
            else:
                assert solver._local_solve_info.explored_nodes_count == 0
                assert not problem._notify_new_best_node_args[1]
        else:
            assert problem._notify_new_best_node_call_count == 0
        best_node_ = Node()
        best_node_.objective = (1 if (sense == minimize) else -1)
        best_node_._uuid = 'abcd'
        problem = DummyProblem(sense)
        results = solver.solve(problem,
                               best_node=best_node_)
        assert results.solution_status == "optimal"
        assert results.termination_condition == "optimality"
        assert results.objective == 0
        assert results.bound == 0
        assert results.absolute_gap == 0
        assert results.relative_gap == 0
        assert results.nodes == 1
        assert results.wall_time is not None
        assert results.best_node._uuid != best_node_._uuid
        if (comm is None) or (comm.size == 1):
            assert results.best_node is not best_node_
        assert results.best_node.objective == results.objective
        assert results.best_node.tree_depth == 0
        if solver.is_worker:
            assert problem._notify_new_best_node_call_count >= 1
            assert problem._notify_new_best_node_args[0] is results.best_node
            if problem._notify_new_best_node_call_count == 2:
                assert problem._notify_new_best_node_args[1]
            else:
                assert problem._notify_new_best_node_args[0]
        else:
            assert problem._notify_new_best_node_call_count == 0

    # empty initial queue
    queue = DispatcherQueueData(
        nodes=[],
        worst_terminal_bound=None,
        sense=minimize)
    for sense in (minimize, maximize):
        queue.sense = sense
        problem = DummyProblem(sense)
        results = solver.solve(problem,
                               initialize_queue=queue)
        assert results.solution_status == "unknown"
        assert results.termination_condition == "no_nodes"
        assert results.objective == (inf if (sense == minimize) else -inf)
        assert results.bound == (-inf if (sense == minimize) else inf)
        assert results.absolute_gap is None
        assert results.relative_gap is None
        assert results.nodes == 0
        assert results.wall_time is not None
        assert results.best_node is None
        assert problem._notify_new_best_node_call_count == 0
        problem = DummyProblem(sense)
        results = solver.solve(problem,
                               initialize_queue=queue,
                               best_objective=0)
        assert results.solution_status == "feasible"
        assert results.termination_condition == "no_nodes"
        assert results.objective == 0
        assert results.bound == (-inf if (sense == minimize) else inf)
        assert results.absolute_gap == inf
        assert results.relative_gap == inf
        assert results.nodes == 0
        assert results.wall_time is not None
        assert results.best_node is None
        assert problem._notify_new_best_node_call_count == 0
        problem = DummyProblem(sense)
        results = solver.solve(problem,
                               initialize_queue=queue,
                               best_objective=0,
                               disable_objective_call=True)
        assert results.solution_status == "feasible"
        assert results.termination_condition == "no_nodes"
        assert results.objective == 0
        assert results.bound == (-inf if (sense == minimize) else inf)
        assert results.absolute_gap == inf
        assert results.relative_gap == inf
        assert results.nodes == 0
        assert results.wall_time is not None
        assert results.best_node is None
        assert problem._notify_new_best_node_call_count == 0
        best_node_ = Node()
        best_node_.objective = (1 if (sense == minimize) else -1)
        best_node_._uuid = 'abcd'
        problem = DummyProblem(sense)
        results = solver.solve(problem,
                               initialize_queue=queue,
                               best_objective=0,
                               best_node=best_node_)
        assert results.solution_status == "feasible"
        assert results.termination_condition == "no_nodes"
        assert results.objective == 0
        assert results.bound == (-inf if (sense == minimize) else inf)
        assert results.absolute_gap == inf
        assert results.relative_gap == inf
        assert results.nodes == 0
        assert results.wall_time is not None
        assert results.best_node._uuid == best_node_._uuid
        if (comm is None) or (comm.size == 1):
            assert results.best_node is best_node_
        if solver.is_worker:
            assert problem._notify_new_best_node_call_count == 1
            assert problem._notify_new_best_node_args[0] is results.best_node
            assert solver._local_solve_info.explored_nodes_count == 0
            assert not problem._notify_new_best_node_args[1]
        else:
            assert problem._notify_new_best_node_call_count == 0
        best_node_ = Node()
        best_node_.objective = (1 if (sense == minimize) else -1)
        best_node_._uuid = 'abcd'
        problem = DummyProblem(sense)
        results = solver.solve(problem,
                               initialize_queue=queue,
                               best_objective=(2 if (sense == minimize) else -2),
                               best_node=best_node_)
        assert results.solution_status == "feasible"
        assert results.termination_condition == "no_nodes"
        assert results.objective == (1 if (sense == minimize) else -1)
        assert results.bound == (-inf if (sense == minimize) else inf)
        assert results.absolute_gap == inf
        assert results.relative_gap == inf
        assert results.nodes == 0
        assert results.wall_time is not None
        assert results.best_node._uuid == best_node_._uuid
        if (comm is None) or (comm.size == 1):
            assert results.best_node is best_node_
        if solver.is_worker:
            assert problem._notify_new_best_node_call_count == 1
            assert problem._notify_new_best_node_args[0] is results.best_node
            assert solver._local_solve_info.explored_nodes_count == 0
            assert not problem._notify_new_best_node_args[1]
        else:
            assert problem._notify_new_best_node_call_count == 0

    # non-empty initial queue
    root = Node()
    root._uuid = 'abcd'
    root.tree_depth = 0
    root.objective = 0
    root.bound = 0
    queue = DispatcherQueueData(
        nodes=[root],
        worst_terminal_bound=None,
        sense=minimize)
    orig_objective = queue.nodes[0].objective
    for sense in (minimize, maximize):
        queue.sense = sense
        queue.nodes[0].objective = orig_objective
        problem = DummyProblem(sense)
        results = solver.solve(problem,
                               initialize_queue=queue)
        assert results.solution_status == "optimal"
        assert results.termination_condition == "optimality"
        assert results.objective == 0
        assert results.bound == 0
        assert results.absolute_gap == 0
        assert results.relative_gap == 0
        assert results.nodes == 1
        assert results.wall_time is not None
        assert results.best_node._uuid == root._uuid
        if (comm is None) or (comm.size == 1):
            assert results.best_node is root
        if solver.is_worker:
            assert problem._notify_new_best_node_call_count == 1
            assert problem._notify_new_best_node_args[0] is results.best_node
            if solver._local_solve_info.explored_nodes_count == 1:
                assert problem._notify_new_best_node_args[1]
            else:
                assert solver._local_solve_info.explored_nodes_count == 0
                assert not problem._notify_new_best_node_args[1]
        else:
            assert problem._notify_new_best_node_call_count == 0
        queue.nodes[0].objective = orig_objective
        problem = DummyProblem(sense)
        results = solver.solve(problem,
                               initialize_queue=queue,
                               best_objective=(1 if (sense == minimize) else -1))
        assert results.solution_status == "optimal"
        assert results.termination_condition == "optimality"
        assert results.objective == 0
        assert results.bound == 0
        assert results.absolute_gap == 0
        assert results.relative_gap == 0
        assert results.nodes == 1
        assert results.wall_time is not None
        assert results.best_node._uuid == root._uuid
        if (comm is None) or (comm.size == 1):
            assert results.best_node is root
        if solver.is_worker:
            assert problem._notify_new_best_node_call_count == 1
            assert problem._notify_new_best_node_args[0] is results.best_node
            if solver._local_solve_info.explored_nodes_count == 1:
                assert problem._notify_new_best_node_args[1]
            else:
                assert solver._local_solve_info.explored_nodes_count == 0
                assert not problem._notify_new_best_node_args[1]
        else:
            assert problem._notify_new_best_node_call_count == 0
        queue.nodes[0].objective = (inf if (sense == minimize) else -inf)
        problem = DummyProblem(sense)
        results = solver.solve(problem,
                               initialize_queue=queue,
                               best_objective=(1 if (sense == minimize) else -1))
        assert results.solution_status == "optimal"
        assert results.termination_condition == "optimality"
        assert results.objective == 0
        assert results.bound == 0
        assert results.absolute_gap == 0
        assert results.relative_gap == 0
        assert results.nodes == 1
        assert results.wall_time is not None
        assert results.best_node._uuid == root._uuid
        if (comm is None) or (comm.size == 1):
            assert results.best_node is root
        if solver.is_worker:
            assert problem._notify_new_best_node_call_count == 1
            assert problem._notify_new_best_node_args[0] is results.best_node
            if solver._local_solve_info.explored_nodes_count == 1:
                assert problem._notify_new_best_node_args[1]
            else:
                assert solver._local_solve_info.explored_nodes_count == 0
                assert not problem._notify_new_best_node_args[1]
        else:
            assert problem._notify_new_best_node_call_count == 0
        queue.nodes[0].objective = (inf if (sense == minimize) else -inf)
        problem = DummyProblem(sense)
        results = solver.solve(problem,
                               initialize_queue=queue,
                               best_objective=(1 if (sense == minimize) else -1),
                               disable_objective_call=True)
        assert results.solution_status == "feasible"
        assert results.termination_condition == "no_nodes"
        assert results.objective == (1 if (sense == minimize) else -1)
        assert results.bound == 0
        assert results.absolute_gap == 1
        assert results.relative_gap == 1
        assert results.nodes == 1
        assert results.wall_time is not None
        assert results.best_node is None
        assert problem._notify_new_best_node_call_count == 0
        queue.nodes[0].objective = orig_objective

def test_initialize_queue_nocomm():
    _test_initialize_queue(None)

def test_solver_nocomm():
    solver = Solver(comm=None)
    assert solver.is_worker
    assert solver.is_dispatcher
    assert solver.comm is None
    assert solver.worker_comm is None

if mpi_available:

    @MPITest(commsize=[1, 2, 3])
    def test_initialize_queue(comm):
        _test_initialize_queue(comm)

    @MPITest(commsize=[1, 2, 3])
    def test_solver_comm(comm):
        solver = Solver(comm=comm, dispatcher_rank=0)
        if comm.size > 1:
            if comm.rank == 0:
                assert solver.is_dispatcher
                assert not solver.is_worker
                assert solver.comm is comm
                assert solver.worker_comm is None
            else:
                assert not solver.is_dispatcher
                assert solver.is_worker
                assert solver.comm is comm
                assert solver.worker_comm is not None
        else:
            assert solver.is_dispatcher
            assert solver.is_worker
            assert solver.comm is comm
            assert solver.worker_comm is comm



import time

from pybnb.common import (minimize,
                          maximize,
                          inf)
from pybnb.problem import (Problem,
                           _SolveInfo,
                           _SimpleSolveInfoCollector)
import pytest

class MinProblem(Problem):
    def sense(self): return minimize

class MaxProblem(Problem):
    def sense(self): return maximize

class TestProblem(object):

    def test_infeasible_objective(self):
        p = MinProblem()
        assert p.infeasible_objective() == inf
        p = MaxProblem()
        assert p.infeasible_objective() == -inf

    def test_unbounded_objective(self):
        p = MinProblem()
        assert p.unbounded_objective() == -inf
        p = MaxProblem()
        assert p.unbounded_objective() == inf

    def test_default_methods(self):
        p = Problem()
        with pytest.raises(NotImplementedError):
            p.sense()
        with pytest.raises(NotImplementedError):
            p.objective()
        with pytest.raises(NotImplementedError):
            p.load_state(None)
        with pytest.raises(NotImplementedError):
            p.save_state(None)
        with pytest.raises(NotImplementedError):
            p.branch()
        p.notify_solve_begins(None, None, None)
        p.notify_new_best_node(None, None)
        p.notify_solve_finished(None, None, None)

class Test_SolveInfo(object):

    def test_methods(self):
        info = _SolveInfo()
        assert info.explored_nodes_count == 0
        assert info.total_queue_time == 0
        assert info.queue_call_count == 0
        assert info.total_objective_time == 0
        assert info.objective_call_count == 0
        assert info.total_bound_time == 0
        assert info.bound_call_count == 0
        assert info.total_branch_time == 0
        assert info.branch_call_count == 0
        assert info.total_load_state_time == 0
        assert info.load_state_call_count == 0
        info.explored_nodes_count += 1
        assert info.explored_nodes_count == 1
        assert info.total_queue_time == 0
        assert info.queue_call_count == 0
        assert info.total_objective_time == 0
        assert info.objective_call_count == 0
        assert info.total_bound_time == 0
        assert info.bound_call_count == 0
        assert info.total_branch_time == 0
        assert info.branch_call_count == 0
        assert info.total_load_state_time == 0
        assert info.load_state_call_count == 0
        info._increment_explored_nodes_stat(2)
        assert info.explored_nodes_count == 3
        assert info.total_queue_time == 0
        assert info.queue_call_count == 0
        assert info.total_objective_time == 0
        assert info.objective_call_count == 0
        assert info.total_bound_time == 0
        assert info.bound_call_count == 0
        assert info.total_branch_time == 0
        assert info.branch_call_count == 0
        assert info.total_load_state_time == 0
        assert info.load_state_call_count == 0
        info.total_queue_time += 1.5
        assert info.explored_nodes_count == 3
        assert info.total_queue_time == 1.5
        assert info.queue_call_count == 0
        assert info.total_objective_time == 0
        assert info.objective_call_count == 0
        assert info.total_bound_time == 0
        assert info.bound_call_count == 0
        assert info.total_branch_time == 0
        assert info.branch_call_count == 0
        assert info.total_load_state_time == 0
        assert info.load_state_call_count == 0
        info.queue_call_count += 1
        assert info.explored_nodes_count == 3
        assert info.total_queue_time == 1.5
        assert info.queue_call_count == 1
        assert info.total_objective_time == 0
        assert info.objective_call_count == 0
        assert info.total_bound_time == 0
        assert info.bound_call_count == 0
        assert info.total_branch_time == 0
        assert info.branch_call_count == 0
        assert info.total_load_state_time == 0
        assert info.load_state_call_count == 0
        info._increment_queue_stat(2.0, 2)
        assert info.explored_nodes_count == 3
        assert info.total_queue_time == 3.5
        assert info.queue_call_count == 3
        assert info.total_objective_time == 0
        assert info.objective_call_count == 0
        assert info.total_bound_time == 0
        assert info.bound_call_count == 0
        assert info.total_branch_time == 0
        assert info.branch_call_count == 0
        assert info.total_load_state_time == 0
        assert info.load_state_call_count == 0
        info.total_objective_time += 1.5
        assert info.explored_nodes_count == 3
        assert info.total_queue_time == 3.5
        assert info.queue_call_count == 3
        assert info.total_objective_time == 1.5
        assert info.objective_call_count == 0
        assert info.total_bound_time == 0
        assert info.bound_call_count == 0
        assert info.total_branch_time == 0
        assert info.branch_call_count == 0
        assert info.total_load_state_time == 0
        assert info.load_state_call_count == 0
        info.objective_call_count += 1
        assert info.explored_nodes_count == 3
        assert info.total_queue_time == 3.5
        assert info.queue_call_count == 3
        assert info.total_objective_time == 1.5
        assert info.objective_call_count == 1
        assert info.total_bound_time == 0
        assert info.bound_call_count == 0
        assert info.total_branch_time == 0
        assert info.branch_call_count == 0
        assert info.total_load_state_time == 0
        assert info.load_state_call_count == 0
        info._increment_objective_stat(2.0, 2)
        assert info.explored_nodes_count == 3
        assert info.total_queue_time == 3.5
        assert info.queue_call_count == 3
        assert info.total_objective_time == 3.5
        assert info.objective_call_count == 3
        assert info.total_bound_time == 0
        assert info.bound_call_count == 0
        assert info.total_branch_time == 0
        assert info.branch_call_count == 0
        assert info.total_load_state_time == 0
        assert info.load_state_call_count == 0
        info.total_bound_time += 1.5
        assert info.explored_nodes_count == 3
        assert info.total_queue_time == 3.5
        assert info.queue_call_count == 3
        assert info.total_objective_time == 3.5
        assert info.objective_call_count == 3
        assert info.total_bound_time == 1.5
        assert info.bound_call_count == 0
        assert info.total_branch_time == 0
        assert info.branch_call_count == 0
        assert info.total_load_state_time == 0
        assert info.load_state_call_count == 0
        info.bound_call_count += 1
        assert info.explored_nodes_count == 3
        assert info.total_queue_time == 3.5
        assert info.queue_call_count == 3
        assert info.total_objective_time == 3.5
        assert info.objective_call_count == 3
        assert info.total_bound_time == 1.5
        assert info.bound_call_count == 1
        assert info.total_branch_time == 0
        assert info.branch_call_count == 0
        assert info.total_load_state_time == 0
        assert info.load_state_call_count == 0
        info._increment_bound_stat(2.0, 2)
        assert info.explored_nodes_count == 3
        assert info.total_queue_time == 3.5
        assert info.queue_call_count == 3
        assert info.total_objective_time == 3.5
        assert info.objective_call_count == 3
        assert info.total_bound_time == 3.5
        assert info.bound_call_count == 3
        assert info.total_branch_time == 0
        assert info.branch_call_count == 0
        assert info.total_load_state_time == 0
        assert info.load_state_call_count == 0
        info.total_branch_time += 1.5
        assert info.explored_nodes_count == 3
        assert info.total_queue_time == 3.5
        assert info.queue_call_count == 3
        assert info.total_objective_time == 3.5
        assert info.objective_call_count == 3
        assert info.total_bound_time == 3.5
        assert info.bound_call_count == 3
        assert info.total_branch_time == 1.5
        assert info.branch_call_count == 0
        assert info.total_load_state_time == 0
        assert info.load_state_call_count == 0
        info.branch_call_count += 1
        assert info.explored_nodes_count == 3
        assert info.total_queue_time == 3.5
        assert info.queue_call_count == 3
        assert info.total_objective_time == 3.5
        assert info.objective_call_count == 3
        assert info.total_bound_time == 3.5
        assert info.bound_call_count == 3
        assert info.total_branch_time == 1.5
        assert info.branch_call_count == 1
        assert info.total_load_state_time == 0
        assert info.load_state_call_count == 0
        info._increment_branch_stat(2.0, 2)
        assert info.explored_nodes_count == 3
        assert info.total_queue_time == 3.5
        assert info.queue_call_count == 3
        assert info.total_objective_time == 3.5
        assert info.objective_call_count == 3
        assert info.total_bound_time == 3.5
        assert info.bound_call_count == 3
        assert info.total_branch_time == 3.5
        assert info.branch_call_count == 3
        assert info.total_load_state_time == 0
        assert info.load_state_call_count == 0
        info.total_load_state_time += 1.5
        assert info.explored_nodes_count == 3
        assert info.total_queue_time == 3.5
        assert info.queue_call_count == 3
        assert info.total_objective_time == 3.5
        assert info.objective_call_count == 3
        assert info.total_bound_time == 3.5
        assert info.bound_call_count == 3
        assert info.total_branch_time == 3.5
        assert info.branch_call_count == 3
        assert info.total_load_state_time == 1.5
        assert info.load_state_call_count == 0
        info.load_state_call_count += 1
        assert info.explored_nodes_count == 3
        assert info.total_queue_time == 3.5
        assert info.queue_call_count == 3
        assert info.total_objective_time == 3.5
        assert info.objective_call_count == 3
        assert info.total_bound_time == 3.5
        assert info.bound_call_count == 3
        assert info.total_branch_time == 3.5
        assert info.branch_call_count == 3
        assert info.total_load_state_time == 1.5
        assert info.load_state_call_count == 1
        info._increment_load_state_stat(2.0, 2)
        assert info.explored_nodes_count == 3
        assert info.total_queue_time == 3.5
        assert info.queue_call_count == 3
        assert info.total_objective_time == 3.5
        assert info.objective_call_count == 3
        assert info.total_bound_time == 3.5
        assert info.bound_call_count == 3
        assert info.total_branch_time == 3.5
        assert info.branch_call_count == 3
        assert info.total_load_state_time == 3.5
        assert info.load_state_call_count == 3
        info.reset()
        assert info.explored_nodes_count == 0
        assert info.total_queue_time == 0
        assert info.queue_call_count == 0
        assert info.total_objective_time == 0
        assert info.objective_call_count == 0
        assert info.total_bound_time == 0
        assert info.bound_call_count == 0
        assert info.total_branch_time == 0
        assert info.branch_call_count == 0
        assert info.total_load_state_time == 0
        assert info.load_state_call_count == 0

    def test_add_from(self):
        info = _SolveInfo()
        info.explored_nodes_count = 1
        info.total_queue_time = 2
        info.queue_call_count = 3
        info.total_objective_time = 4
        info.objective_call_count = 5
        info.total_bound_time = 6
        info.bound_call_count = 7
        info.total_branch_time = 8
        info.branch_call_count = 9
        info.total_load_state_time = 10
        info.load_state_call_count = 11
        info1 = _SolveInfo()
        info1.explored_nodes_count = -2
        info1.total_queue_time = -3
        info1.queue_call_count = -4
        info1.total_objective_time = -5
        info1.objective_call_count = -6
        info1.total_bound_time = -7
        info1.bound_call_count = -8
        info1.total_branch_time = -9
        info1.branch_call_count = -10
        info1.total_load_state_time = -11
        info1.load_state_call_count = -12
        info.add_from(info1)
        assert info.explored_nodes_count  == -1
        assert info.total_queue_time  == -1
        assert info.queue_call_count  == -1
        assert info.total_objective_time  == -1
        assert info.objective_call_count  == -1
        assert info.total_bound_time  == -1
        assert info.bound_call_count  == -1
        assert info.total_branch_time  == -1
        assert info.branch_call_count  == -1
        assert info.total_load_state_time  == -1
        assert info.load_state_call_count  == -1
        with pytest.raises(TypeError):
            info.add_from(1)

    def test__SimpleSolveInfoCollector(self):
        class Junk(Problem):
            def __init__(self):
                self.d = {}
                self.d['sense'] = False
                self.d['objective'] = False
                self.d['bound'] = False
                self.d['save_state'] = False
                self.d['load_state'] = False
                self.d['branch'] = False
                self.d['notify_solve_begins'] = False
                self.d['notify_new_best_node'] = False
                self.d['notify_solve_finished'] = False
            def sense(self):
                self.d['sense'] = True
                return -1
            def objective(self):
                time.sleep(0.01)
                self.d['objective'] = True
                return -2
            def bound(self):
                time.sleep(0.01)
                self.d['bound'] = True
                return -3
            def save_state(self, node):
                self.d['save_state'] = True
            def load_state(self, node):
                time.sleep(0.01)
                self.d['load_state'] = True
            def branch(self):
                time.sleep(0.01)
                self.d['branch'] = True
                return ()
            def notify_solve_begins(self,
                                      comm,
                                      worker_comm,
                                      convergence_checker):
                self.d['notify_solve_begins'] = True
            def notify_new_best_node(self,
                                     node,
                                     current):
                self.d['notify_new_best_node'] = True
            def notify_solve_finished(self,
                                      comm,
                                      worker_comm,
                                      results):
                self.d['notify_solve_finished'] = True
        j = Junk()
        assert j.d['sense'] == False
        assert j.d['objective'] == False
        assert j.d['bound'] == False
        assert j.d['save_state'] == False
        assert j.d['load_state'] == False
        assert j.d['branch'] == False
        assert j.d['notify_solve_begins'] == False
        assert j.d['notify_new_best_node'] == False
        assert j.d['notify_solve_finished'] == False
        p = _SimpleSolveInfoCollector(j)
        p.set_clock(time.time)
        info = _SolveInfo()
        assert info.explored_nodes_count == 0
        assert info.total_queue_time == 0
        assert info.queue_call_count == 0
        assert info.total_objective_time == 0
        assert info.objective_call_count == 0
        assert info.total_bound_time == 0
        assert info.bound_call_count == 0
        assert info.total_branch_time == 0
        assert info.branch_call_count == 0
        p.set_solve_info_object(info)
        assert j.d['sense'] == False
        assert j.d['objective'] == False
        assert j.d['bound'] == False
        assert j.d['save_state'] == False
        assert j.d['load_state'] == False
        assert j.d['branch'] == False
        assert j.d['notify_solve_begins'] == False
        assert j.d['notify_new_best_node'] == False
        assert j.d['notify_solve_finished'] == False
        assert info.explored_nodes_count == 0
        assert info.total_queue_time == 0
        assert info.queue_call_count == 0
        assert info.total_objective_time == 0
        assert info.objective_call_count == 0
        assert info.total_bound_time == 0
        assert info.bound_call_count == 0
        assert info.total_branch_time == 0
        assert info.branch_call_count == 0
        assert info.total_load_state_time == 0
        assert info.load_state_call_count == 0
        p.sense()
        assert j.d['sense'] == True
        assert j.d['objective'] == False
        assert j.d['bound'] == False
        assert j.d['save_state'] == False
        assert j.d['load_state'] == False
        assert j.d['branch'] == False
        assert j.d['notify_solve_begins'] == False
        assert j.d['notify_new_best_node'] == False
        assert j.d['notify_solve_finished'] == False
        assert info.explored_nodes_count == 0
        assert info.total_queue_time == 0
        assert info.queue_call_count == 0
        assert info.total_objective_time == 0
        assert info.objective_call_count == 0
        assert info.total_bound_time == 0
        assert info.bound_call_count == 0
        assert info.total_branch_time == 0
        assert info.branch_call_count == 0
        assert info.total_load_state_time == 0
        assert info.load_state_call_count == 0
        p.objective()
        assert j.d['sense'] == True
        assert j.d['objective'] == True
        assert j.d['bound'] == False
        assert j.d['save_state'] == False
        assert j.d['load_state'] == False
        assert j.d['branch'] == False
        assert j.d['notify_solve_begins'] == False
        assert j.d['notify_new_best_node'] == False
        assert j.d['notify_solve_finished'] == False
        assert info.explored_nodes_count == 0
        assert info.total_queue_time == 0
        assert info.queue_call_count == 0
        assert info.total_objective_time > 0
        assert info.objective_call_count == 1
        assert info.total_bound_time == 0
        assert info.bound_call_count == 0
        assert info.total_branch_time == 0
        assert info.branch_call_count == 0
        assert info.total_load_state_time == 0
        assert info.load_state_call_count == 0
        p.bound()
        assert j.d['sense'] == True
        assert j.d['objective'] == True
        assert j.d['bound'] == True
        assert j.d['save_state'] == False
        assert j.d['load_state'] == False
        assert j.d['branch'] == False
        assert j.d['notify_solve_begins'] == False
        assert j.d['notify_new_best_node'] == False
        assert j.d['notify_solve_finished'] == False
        assert info.explored_nodes_count == 0
        assert info.total_queue_time == 0
        assert info.queue_call_count == 0
        assert info.total_objective_time > 0
        assert info.objective_call_count == 1
        assert info.total_bound_time > 0
        assert info.bound_call_count == 1
        assert info.total_branch_time == 0
        assert info.branch_call_count == 0
        assert info.total_load_state_time == 0
        assert info.load_state_call_count == 0
        p.save_state(None)
        assert j.d['sense'] == True
        assert j.d['objective'] == True
        assert j.d['bound'] == True
        assert j.d['save_state'] == True
        assert j.d['load_state'] == False
        assert j.d['branch'] == False
        assert j.d['notify_solve_begins'] == False
        assert j.d['notify_new_best_node'] == False
        assert j.d['notify_solve_finished'] == False
        assert info.explored_nodes_count == 0
        assert info.total_queue_time == 0
        assert info.queue_call_count == 0
        assert info.total_objective_time > 0
        assert info.objective_call_count == 1
        assert info.total_bound_time > 0
        assert info.bound_call_count == 1
        assert info.total_branch_time == 0
        assert info.branch_call_count == 0
        assert info.total_load_state_time == 0
        assert info.load_state_call_count == 0
        p.load_state(None)
        assert j.d['sense'] == True
        assert j.d['objective'] == True
        assert j.d['bound'] == True
        assert j.d['save_state'] == True
        assert j.d['load_state'] == True
        assert j.d['branch'] == False
        assert j.d['notify_solve_begins'] == False
        assert j.d['notify_new_best_node'] == False
        assert j.d['notify_solve_finished'] == False
        assert info.explored_nodes_count == 0
        assert info.total_queue_time == 0
        assert info.queue_call_count == 0
        assert info.total_objective_time > 0
        assert info.objective_call_count == 1
        assert info.total_bound_time > 0
        assert info.bound_call_count == 1
        assert info.total_branch_time == 0
        assert info.branch_call_count == 0
        assert info.total_load_state_time > 0
        assert info.load_state_call_count == 1
        list(p.branch())
        assert j.d['sense'] == True
        assert j.d['objective'] == True
        assert j.d['bound'] == True
        assert j.d['save_state'] == True
        assert j.d['load_state'] == True
        assert j.d['branch'] == True
        assert j.d['notify_solve_begins'] == False
        assert j.d['notify_new_best_node'] == False
        assert j.d['notify_solve_finished'] == False
        assert info.explored_nodes_count == 0
        assert info.total_queue_time == 0
        assert info.queue_call_count == 0
        assert info.total_objective_time > 0
        assert info.objective_call_count == 1
        assert info.total_bound_time > 0
        assert info.bound_call_count == 1
        assert info.total_branch_time > 0
        assert info.branch_call_count == 1
        assert info.total_load_state_time > 0
        assert info.load_state_call_count == 1
        p.notify_solve_begins(None, None, None)
        assert j.d['sense'] == True
        assert j.d['objective'] == True
        assert j.d['bound'] == True
        assert j.d['save_state'] == True
        assert j.d['load_state'] == True
        assert j.d['branch'] == True
        assert j.d['notify_solve_begins'] == True
        assert j.d['notify_new_best_node'] == False
        assert j.d['notify_solve_finished'] == False
        assert info.explored_nodes_count == 0
        assert info.total_queue_time == 0
        assert info.queue_call_count == 0
        assert info.total_objective_time > 0
        assert info.objective_call_count == 1
        assert info.total_bound_time > 0
        assert info.bound_call_count == 1
        assert info.total_branch_time > 0
        assert info.branch_call_count == 1
        assert info.total_load_state_time > 0
        assert info.load_state_call_count == 1
        p.notify_new_best_node(None, None)
        assert j.d['sense'] == True
        assert j.d['objective'] == True
        assert j.d['bound'] == True
        assert j.d['save_state'] == True
        assert j.d['load_state'] == True
        assert j.d['branch'] == True
        assert j.d['notify_solve_begins'] == True
        assert j.d['notify_new_best_node'] == True
        assert j.d['notify_solve_finished'] == False
        assert info.explored_nodes_count == 0
        assert info.total_queue_time == 0
        assert info.queue_call_count == 0
        assert info.total_objective_time > 0
        assert info.objective_call_count == 1
        assert info.total_bound_time > 0
        assert info.bound_call_count == 1
        assert info.total_branch_time > 0
        assert info.branch_call_count == 1
        assert info.total_load_state_time > 0
        assert info.load_state_call_count == 1
        p.notify_solve_finished(None, None, None)
        assert j.d['sense'] == True
        assert j.d['objective'] == True
        assert j.d['bound'] == True
        assert j.d['save_state'] == True
        assert j.d['load_state'] == True
        assert j.d['branch'] == True
        assert j.d['notify_solve_begins'] == True
        assert j.d['notify_new_best_node'] == True
        assert j.d['notify_solve_finished'] == True
        assert info.explored_nodes_count == 0
        assert info.total_queue_time == 0
        assert info.queue_call_count == 0
        assert info.total_objective_time > 0
        assert info.objective_call_count == 1
        assert info.total_bound_time > 0
        assert info.bound_call_count == 1
        assert info.total_branch_time > 0
        assert info.branch_call_count == 1
        assert info.total_load_state_time > 0
        assert info.load_state_call_count == 1
